#!/usr/bin/env python

#################################################################
#
#    shared_vars
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################

#################################################################
#
#   SpliceAI delta scores fields
#
#################################################################
DStags = {'DS_AG', 'DS_AL', 'DS_DG', 'DS_DL'}


#################################################################
#
#   VEP terms corresponding to intronic, intergenic or
#       regulatory regions to be removed or ignored
#
#################################################################
VEPremove = {
            # intronic and intergenic features tags
            'intron_variant', 'intergenic_variant',
            'downstream_gene_variant', 'upstream_gene_variant',
            'NMD_transcript_variant', 'non_coding_transcript_variant',
            'non_coding_transcript_exon_variant',
            # regulatory features tags
            'feature_elongation', 'feature_truncation',
            'regulatory_region_variant', 'regulatory_region_amplification',
            'regulatory_region_ablation', 'splice_region_variant',
            'TFBS_amplification', 'TFBS_ablation', 'TF_binding_site_variant'
            }


#################################################################
#
#   VEP terms to rescue only for SpliceAI variants
#
#################################################################
VEPSpliceAI = {'intron_variant'}


#################################################################
#
#   Unbalanced chromosomes to ignore in novoCaller, PP -> NA
#
#################################################################
real_NA_chroms = {'M', 'MT', 'X', 'Y'}

# for automated tests
test_NA_chroms = {'2'}


#################################################################
#
#   VEP Consequence to impact
#
#################################################################
VEP_encode = {
                # HIGH
                'transcript_ablation': 1,
                'splice_acceptor_variant': 1,
                'splice_donor_variant': 1,
                'stop_gained': 1,
                'frameshift_variant': 1,
                'stop_lost': 1,
                'start_lost': 1,
                'transcript_amplification': 1,
                # MODERATE
                'inframe_insertion': 2,
                'inframe_deletion': 2,
                'missense_variant': 2,
                'protein_altering_variant': 2,
                # LOW
                'splice_region_variant': 3,
                'incomplete_terminal_codon_variant': 3,
                'start_retained_variant': 3,
                'stop_retained_variant': 3,
                'synonymous_variant': 3,
                # MODIFIER
                'MODIFIER': 4
            }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#################################################################
#
#    vcf_parser
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import re
import gzip


#################################################################
#
#    Vcf
#      -> Header
#      -> Variant
#
#################################################################
class Vcf(object):
    ''' object to read and manipulate vcf file format '''

    def __init__(self, inputfile):
        ''' open input vcf, read header lines and save
        information as Header object to initialize Vcf object '''
        self.inputfile = inputfile
        self.header = self.parse_header()
    #end def

    class Header(object):
        ''' object to store vcf header information '''

        def __init__(self, definitions, columns, IDs_genotypes):
            ''' initialize Header object '''
            self.definitions = definitions
            self.columns = columns
            self.IDs_genotypes = IDs_genotypes
        #end def

        def add_tag_definition(self, tag_definition, tag_type='INFO'):
            ''' add tag_definition to the header on top
            of the block specified by tag_type (e.g. FORMAT, INFO) '''
            added_tag, new_definitions = False, ''
            for line in self.definitions.split('\n')[:-1]:
                if line.startswith('##' + tag_type) and not added_tag:
                    added_tag = True
                    new_definitions += tag_definition + '\n'
                #end if
                new_definitions += line + '\n'
            #end for
            self.definitions = new_definitions
        #end def

        def remove_tag_definition(self, tag, tag_type='INFO'):
            ''' remove tag definition from header,
            block specified by tag_type (e.g. FORMAT, INFO) '''
            new_definitions = ''
            for line in self.definitions.split('\n')[:-1]:
                if line.startswith('##' + tag_type + '=<ID=' + tag + ','): ##<tag_type>=<ID=<tag>,...
                    continue
                #end if
                new_definitions += line + '\n'
            #end for
            self.definitions = new_definitions
        #end def

        def get_tag_field_idx(self, tag, field, tag_type='INFO', sep='|'):
            ''' get idx for value field in tag from definition,
            block specified by tag_type (e.g. FORMAT, INFO) '''
            for line in self.definitions.split('\n')[:-1]:
                if line.startswith('##' + tag_type + '=<ID=' + tag + ','):
                    try:
                        format = line.split('Format:')[1]
                        # Cleaning format
                        format = format.replace('\'', '')
                        format = format.replace('\"', '')
                        format = format.replace('>', '')
                    except Exception:
                        raise ValueError('\nERROR in VCF header structure, {0} tag definition has no format specification\n'
                                            .format(tag))
                    #end try
                    # Search exact match
                    # if not exact match, search for partial match (included in field name)
                    for i, field_i in enumerate(format.split(sep)):
                        if field == field_i.strip(): return i # exact match
                        #end if
                    #end for
                    for i, field_i in enumerate(format.split(sep)):
                        if field in field_i.strip(): return i # partial match
                        #end if
                    #end for
                #end if
            #end for
            raise ValueError('\nERROR in VCF header structure, {0} tag definition is missing\n'
                                .format(tag))
        #end def

        def check_tag_definition(self, tag, tag_type='INFO', sep='|'):
            ''' check if tag is standalone or field of another leading tag,
            return leading tag and field index, if any, to acces requested tag '''
            for line in self.definitions.split('\n')[:-1]:
                if line.startswith('##' + tag_type):
                    if ('=<ID=' + tag + ',') in line: ##<tag_type>=<ID=<tag>,..
                        # tag is already a standalone tag
                        return tag, 0
                    elif tag in line and 'Format:' in line: ##<tag_type>=<ID=<lead_tag>,...,Description="... Format:<tag>">
                        # tag is a field, get leading tag and field index
                        lead_tag = line.split('=<ID=')[1].split(',')[0]
                        idx = self.get_tag_field_idx(lead_tag, tag, tag_type, sep)
                        return lead_tag, idx
                    #end if
                #end if
            #end for
            raise ValueError('\nERROR in VCF header structure, {0} tag definition is missing\n'
                                .format(tag))
        #end def

    #end class Header

    class Variant(object):
        ''' object to store information for variant in vcf format '''

        def __init__(self, line_strip, IDs_genotypes):
            ''' initialize Variant object '''
            line_split = line_strip.split('\t')
            self.CHROM = line_split[0]
            self.POS = int(line_split[1])
            self.ID = line_split[2]
            self.REF = line_split[3]
            self.ALT = line_split[4]
            self.QUAL = line_split[5]
            self.FILTER = line_split[6]
            self.INFO = line_split[7]
            if IDs_genotypes: # if samples
                self.FORMAT = line_split[8] # get FORMAT column
            else: self.FORMAT = ''
            #end if
            self.IDs_genotypes = IDs_genotypes
            self.GENOTYPES = {k: v for k, v in zip(IDs_genotypes, line_split[9:])}
        #end def

        def to_string(self):
            ''' variant as string rapresentation '''
            variant_as_list = [ self.CHROM,
                                str(self.POS),
                                self.ID,
                                self.REF,
                                self.ALT,
                                self.QUAL,
                                self.FILTER,
                                self.INFO ]
            if self.IDs_genotypes: # if samples
                variant_as_list.append(self.FORMAT) # add FORMAT column
                for IDs_genotype in self.IDs_genotypes: # add sample columns
                    variant_as_list.append(self.GENOTYPES[IDs_genotype])
                #end for
            #end if

            return '\t'.join(variant_as_list) + '\n'
        #end def

        def repr(self):
            ''' variant representation as CHROM:POSREF>ALT'''
            return '{0}:{1}{2}>{3}'.format(self.CHROM,
                                      self.POS,
                                      self.REF,
                                      self.ALT)
        #end def

        def remove_tag_genotype(self, tag_to_remove, sep=':'):
            ''' remove tag field from FORMAT and GENOTYPES '''
            idx_tag_to_remove, new_format = -1, []
            # Removing tag field from FORMAT
            for i, tag in enumerate(self.FORMAT.split(sep)):
                if tag_to_remove == tag:
                    idx_tag_to_remove = i
                else:
                    new_format.append(tag)
                #end if
            #end for
            # Error if tag_to_remove not found in FORMAT
            if idx_tag_to_remove == -1:
                raise ValueError('\nERROR in variant FORMAT field, {0} tag is missing\n'
                            .format(tag_to_remove))
            #end if
            # Updating FORMAT
            self.FORMAT = sep.join(new_format)
            # Removing tag field from GENOTYPES
            for ID_genotype, genotype in self.GENOTYPES.items():
                genotype_as_list = genotype.split(sep)
                try:
                    del genotype_as_list[idx_tag_to_remove]
                except Exception: # del will fail for trailing fields that are dropped
                                  # field to remove is missing already
                    pass
                #end try
                self.GENOTYPES[ID_genotype] = sep.join(genotype_as_list)
            #end for
        #end def

        def complete_genotype(self, sep=':'):
            ''' fill the trailing fields dropped in GENOTYPES,
            based on FORMAT structure '''
            len_FORMAT = len(self.FORMAT.split(sep))
            for ID_genotype, genotype in self.GENOTYPES.items():
                genotype_as_list = genotype.split(sep)
                for i in range(len_FORMAT - len(genotype_as_list)):
                    genotype_as_list.append('.')
                #end for
                self.GENOTYPES[ID_genotype] = sep.join(genotype_as_list)
            #end for
        #end def

        def empty_genotype(self, sep=':'):
            ''' return a empty genotype based on FORMAT structure '''
            len_FORMAT = len(self.FORMAT.split(sep))
            return './.' + (sep + '.') * (len_FORMAT - 1)
        #end def

        def remove_tag_info(self, tag_to_remove, sep=';'):
            ''' remove tag field from INFO '''
            new_INFO = []
            for tag in self.INFO.split(sep):
                if tag.startswith(tag_to_remove + '='):
                    continue
                #end if
                new_INFO.append(tag)
            #end for
            self.INFO = sep.join(new_INFO)
        #end def

        def add_tag_format(self, tag_to_add, sep=':'):
            ''' add tag field to FORMAT '''
            self.FORMAT += sep + tag_to_add
        #end def

        def add_values_genotype(self, ID_genotype, values, sep=':'):
            ''' add values field to genotype specified by corresponding ID '''
            try:
                self.GENOTYPES[ID_genotype] += sep + values
            except Exception:
                raise ValueError('\nERROR in GENOTYPES identifiers, {0} identifier is missing in VCF\n'
                            .format(ID_genotype))
            #end try
        #end def

        def add_tag_info(self, tag_to_add, sep=';'):
            ''' add tag field and value (tag_to_add) to INFO '''
            # tag_to_add -> tag=<value>
            if self.INFO.endswith(sep): # if INFO ending is wrongly formatted
                self.INFO += tag_to_add
            else:
                self.INFO += sep + tag_to_add
            #end if
        #end def

        def get_tag_value(self, tag_to_get, sep=';'):
            ''' get value from tag (tag_to_get) in INFO '''
            for tag in self.INFO.split(sep):
                if tag.startswith(tag_to_get + '='):
                    try:
                        return tag.split(tag_to_get + '=')[1]
                    except Exception: # tag field is in a wrong format
                        raise ValueError('\nERROR in variant INFO field, {0} tag is in the wrong format\n'
                                    .format(tag_to_get))
                    #end try
                #end if
            #end for

            # tag_to_get not found
            raise ValueError('\nERROR in variant INFO field, {0} tag is missing\n'.format(tag_to_get))
        #end def

        def get_genotype_value(self, ID_genotype, tag_to_get, sep=':'):
            ''' get value from tag (tag_to_get) in genotype specified by corresponding ID '''
            # Get index from FORMAT
            idx_tag_to_get = -1
            for i, tag in enumerate(self.FORMAT.split(sep)):
                if tag_to_get == tag:
                    idx_tag_to_get = i
                    break
                #end if
            #end for
            # Error if tag_to_get not found in FORMAT
            if idx_tag_to_get == -1:
                raise ValueError('\nERROR in variant FORMAT field, {0} tag is missing\n'
                            .format(tag_to_get))
            #end if
            # Get value from index in genotype by ID
            try:
                return self.GENOTYPES[ID_genotype].split(sep)[idx_tag_to_get]
            except Exception:
                raise ValueError('\nERROR in GENOTYPES identifiers, {0} identifier is missing in VCF\n'
                            .format(ID_genotype))
            #end try
        #end def

    #end class Variant

    @staticmethod
    def read_vcf(inputfile):
        ''' read vcf file, gzipped or ungzipped,
        return a generator '''
        if inputfile.endswith('.gz') or \
           inputfile.endswith('.bgz'):
            with gzip.open(inputfile, 'rb') as fz:
                for byteline in fz:
                    yield byteline.decode('utf-8')
                #end for
            #end with
        else:
            with open(inputfile, encoding='utf-8') as fi:
                for line in fi:
                    yield line
                #end for
            #end with
        #end if
    #end def

    def parse_header(self):
        ''' read header and save information as Header object '''
        definitions, columns, IDs_genotypes = '', '', []
        for line in self.read_vcf(self.inputfile):
            if line.startswith('#'): # reading a header line
                line_strip = line.rstrip()
                if line_strip.startswith('##'): # header definition line
                    definitions += line_strip + '\n'
                elif line_strip.startswith('#CHROM'): # header columns line
                    columns += line_strip + '\n'
                    IDs_genotypes = line_strip.split('\t')[9:]
                #end if
            else: # finished to read the header
                break # exit and close buffer
            #end if
        #end for

        # Checking header is correct
        if definitions and columns:
            return self.Header(definitions, columns, IDs_genotypes)
        else:
            raise ValueError('\nERROR in VCF header structure, missing essential lines\n')
        #end if
    #end def

    def parse_variants(self): # generator
        ''' return a generator to variants stored as Variant objects '''
        for line in self.read_vcf(self.inputfile):
            if not line.startswith('#'):
                line_strip = line.rstrip()
                if line_strip:
                    try:
                        yield self.Variant(line_strip, self.header.IDs_genotypes)
                    except Exception:
                        raise ValueError('\nERROR in variant VCF structure, missing essential columns\n')
                    #end try
                #end if
            #end if
        #end for
    #end def

    def write_definitions(self, outputfile_obj):
        ''' write header definitions to outputfile_obj buffer '''
        outputfile_obj.write(self.header.definitions)
    #end def

    def write_columns(self, outputfile_obj):
        ''' write header columns to outputfile_obj buffer,
        #CHROM ... '''
        outputfile_obj.write(self.header.columns)
    #end def

    def write_header(self, outputfile_obj):
        ''' write header definitions and columns to outputfile_obj buffer '''
        self.write_definitions(outputfile_obj)
        self.write_columns(outputfile_obj)
    #end def

    def write_variant(self, outputfile_obj, parsed_variant):
        ''' write parsed_variant (Variant object generated by parse_variants) to outputfile_obj buffer '''
        outputfile_obj.write(parsed_variant.to_string())
    #end def

#end class Vcf



#!/usr/bin/env python

#################################################################
#
#    blackList
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
# shared_functions as *
from granite.lib.shared_functions import *
# vcf_parser
from granite.lib import vcf_parser


#################################################################
#
#    FUNCTIONS
#
#################################################################
#################################################################
#    runner
#################################################################
def main(args):
    ''' '''
    # Variables
    afthr, aftag = 0., ''
    big_dict = {}
    is_afthr = True if args['afthr'] else False
    is_bigfile = True if args['bigfile'] else False
    is_verbose = True if args['verbose'] else False

    # Creating Vcf object
    vcf_obj = vcf_parser.Vcf(args['inputfile'])

    # Check arguments
    if is_afthr:
        afthr = float(args['afthr'])
        if args['aftag']:
            aftag, aftag_idx = vcf_obj.header.check_tag_definition(args['aftag'])
        else:
            sys.exit('\nERROR in parsing arguments: to filter by population allele frequency please specify the TAG to use\n')
        #end if
    else:
        if not is_bigfile:
            sys.exit('\nERROR in parsing arguments: to blacklist specify a BIG file and/or a threshold for population allele frequency and the TAG to use\n')
        #end if
    #end if

    # Buffers
    fo = open(args['outputfile'], 'w')

    # Loading big if specified
    if is_bigfile: big_dict = load_big(args['bigfile'])
    #end if

    # Writing header
    vcf_obj.write_header(fo)

    # Reading variants and writing passed
    analyzed = 0
    for i, vnt_obj in enumerate(vcf_obj.parse_variants()):
        if is_verbose:
            sys.stderr.write('\rAnalyzing variant... ' + str(i + 1))
            sys.stderr.flush()
        #end if

        # # Check if chromosome is canonical and in valid format
        # if not check_chrom(vnt_obj.CHROM):
        #     continue
        # #end if
        analyzed += 1

        # Get allele frequency from aftag tag if requested
        if is_afthr:
            af = allele_frequency(vnt_obj, aftag, aftag_idx)
            # Check allele frequency
            if af > afthr:
                continue
            #end if
        #end if

        if is_bigfile:
            vtype = variant_type(vnt_obj.REF, vnt_obj.ALT)
            try:
                key = vnt_obj.CHROM + '_' + vtype
                is_blacklist = big_dict[key][vnt_obj.POS]
            except Exception:
                sys.exit('\nERROR in blacklist check: {0}:{1} missing in BIG file'.format(key, vnt_obj.POS))
            #end try
            if is_blacklist:
                continue
            #end if
        #end if

        # All good, pass and write variant
        vcf_obj.write_variant(fo, vnt_obj)
    #end for
    sys.stderr.write('\n\n...Wrote results for ' + str(analyzed) + ' analyzed variants out of ' + str(i + 1) + ' total variants\n')
    sys.stderr.flush()

    # Closing buffers
    fo.close()
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python


######################################################################################
##	fasta_parser.py
#
#	Author: Michele Berselli
#		University of Padova
#		berselli.michele@gmail.com
#
##	LICENSE:
#		Copyright (C) 2018  Michele Berselli
#
#		This program is free software: you can redistribute it and/or modify
#		it under the terms of the GNU General Public License as published by
#		the Free Software Foundation.
#
#		This program is distributed in the hope that it will be useful,
#		but WITHOUT ANY WARRANTY; without even the implied warranty of
#		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#		GNU General Public License for more details.
#
#		You should have received a copy of the GNU General Public License
#		along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
######################################################################################


import sys, argparse, os
from bitarray import bitarray


########################################
####		CLASS FastaHandler     ####
########################################
class FastaHandler(object):
	''' a class to implement an object to handle multiple fasta sequences '''

	########################################
	####		CLASS FastaSequence    ####
	########################################
	class FastaSequence(object):
		''' a class to implement an object representing a fasta sequence '''

		## FUNCTIONS ##
		def __init__(self, header, sequence):
			self.__header = header
			self.__sequence = sequence
		#end def __init__

		def get_header(self):
			return self.__header
		#end def get_header

		def get_sequence(self):
			if type(self.__sequence) == bitarray:
				return self.__bitarray_to_seq()
			else:
				return self.__sequence
			#end if
		#end def get_seq()

		def get_bitarray(self):
			if type(self.__sequence) != bitarray:
				raise ValueError('Sequence is not encoded as a bitarray, use get_sequence() instead!')
			else:
				return self.__sequence
			#end if
		#end def get_bitarray

		def __bitarray_to_seq(self):
			s, i, max_i = '', 0, len(self.__sequence) - 2
			encode = {'00': 'A', '01': 'C', '10': 'G', '11': 'T'}
			while i <= max_i:
				s += encode[str(int(self.__sequence[i])) + str(int(self.__sequence[i + 1]))]
				i += 2
			#end while
			return s
		#end def __bitarray_to_seq

		def write_sequence(self, fo, max_char_per_line=0):
			fo.write('>' + self.__header + '\n')
			i, sequence = 0, self.get_sequence()
			if not max_char_per_line:
				fo.write(sequence + '\n')
			else:
				max_i = len(sequence) - max_char_per_line
				while i <= max_i:
					fo.write(sequence[i:i + max_char_per_line] + '\n')
					i += max_char_per_line
				#end while
				if sequence[i:]: fo.write(sequence[i:] + '\n')
			#end if
		#def write_sequence

	#end class FastaSequence

	## FUNCTIONS ##
	def __init__(self):
		self.__sequences = []
	#end def __init__

	def parse(self, inputfile):
		''' adds fasta sequences as FastaSequence objects '''
		header, sequence = None, []
		fi = self.__open(inputfile)
		for line in fi:
			if line.startswith('>'):
				if header: self.__sequences.append(self.FastaSequence(header, ''.join(sequence)))
				header, sequence = line.rstrip()[1:], []
			else:
				sequence.append(line.rstrip())
			#end if
		#end for
		if header: self.__sequences.append(self.FastaSequence(header, ''.join(sequence)))
		fi.close()
	#end def parse

	def parse_binary(self, inputfile):
		''' adds fasta sequences binary encoded as FastaSequence objects, 
		works only for upper or lower canonical bases A, C, T, G '''
		encode = {
				'A': '00', 'a': '00',
				'C': '01', 'c': '01',
				'G': '10', 'g': '10',
				'T': '11', 't': '11'
				}
		header, sequence = None, bitarray()
		fi = self.__open(inputfile)
		for line in fi:
			if line.startswith('>'):
				if header: self.__sequences.append(self.FastaSequence(header, sequence))
				header, sequence = line.rstrip()[1:], bitarray()
			else:
				try:
					for c in line.rstrip(): sequence.extend(encode[c])
				except:
					self.__sequences = []
					raise ValueError('Non-canonical base found in sequence ' + header + '!')
				#end try
			#end if
		#end for
		if header: self.__sequences.append(self.FastaSequence(header, sequence))
		fi.close()
	#end def parse_binary

	def parse_generator(self, inputfile):
		''' create a generator of the sequences, 
		each iteration return a (header, sequence) tuple '''
		header, sequence = None, []
		fi = self.__open(inputfile)
		for line in fi:
			if line.startswith('>'):
				if header: yield (header, ''.join(sequence))
				header, sequence = line.rstrip()[1:], []
			else:
				sequence.append(line.rstrip())
			#end if
		#end for
		if header: yield (header, ''.join(sequence))
		fi.close()
	#end def parse_generator

	def __open(self, inputfile):
		''' check the existance of inputfile and opens it if exists '''
		if os.path.isfile(inputfile):
			return open(inputfile, 'r')
		else:
			raise ValueError('Input file is missing!')
		#end if
	#end def __open

	def iter_sequences(self):
		''' create a generator of the stored sequences,
		each iteration return a FastaSequence object '''
		return iter(self.__sequences)
	#end def iter_sequences

	def get_sequences(self):
		''' return the list of the stored FastaSequence objects '''
		return self.__sequences
	#end def get_sequences

	def write_sequences(self, fo, max_char_per_line=0):
		for s in self.__sequences:
			s.write_sequence(fo, max_char_per_line)
		#end for
	#end def write_sequences

#end class FastaHandler




#!/usr/bin/env python

#################################################################
#
#    toPED
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import json
# pedigree_parser
from granite.lib import pedigree_parser


#################################################################
#
#    FUNCTIONS
#
#################################################################
#################################################################
#    runner
#################################################################
def main(args):
    ''' '''
    # Variables
    encode_gender = {'M': 1, 'F': 2}
    familyID = args['family'] if args['family'] else 'FAM'

    # Loading pedigree
    if os.path.isfile(args['pedigree']):
        with open(args['pedigree']) as fi:
            pedigree = json.load(fi)
        #end with
    else:
        try: pedigree = json.loads(args['pedigree'])
        except Exception:
            sys.exit('\nERROR in parsing arguments: pedigree must be either a json file or a string representing a json\n')
        #end try
    #end if

    # Creating Pedigree object
    pedigree_obj = pedigree_parser.Pedigree(pedigree)

    # Buffers
    fo = open(args['outputfile'], 'w')

    # Writing PED
    for _, member_obj in pedigree_obj.members.items():
        if member_obj.is_sample():
            memberID = member_obj.sample
            try: sex = encode_gender[member_obj.gender]
            except Exception: sex = 'U'
            #end try
            paternalID, maternalID = 0, 0
            if member_obj.has_parents():
                for parent in member_obj.get_parents():
                    if parent.gender == 'M':
                        paternalID = parent.sample
                    elif parent.gender == 'F':
                        maternalID = parent.sample
                    #end if
                #end for
            #end if
            fo.write('{0}\t{1}\t{2}\t{3}\t{4}\t-9\n'.format(
                                                    familyID,
                                                    memberID,
                                                    paternalID,
                                                    maternalID,
                                                    sex
                                                    ))
        #end if
    #end for

    # Closing buffers
    fo.close()
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python

#################################################################
#
#    novoCaller
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import pysam
import tabix
import numpy as np
# shared_functions as *
from granite.lib.shared_functions import *
# vcf_parser
from granite.lib import vcf_parser
# shared_vars
from granite.lib.shared_vars import real_NA_chroms
from granite.lib.shared_vars import test_NA_chroms

# suppress warnings
import warnings
warnings.filterwarnings("ignore")


#################################################################
#
#    FUNCTIONS
#
#################################################################
#################################################################
#    GT_ordering_alternate (original)
#################################################################
def GT_ordering_alternate(ALT_count):
    ''' '''
    combos = (ALT_count + 1) * (ALT_count + 2) // 2
    ordering = np.empty([combos, 2])
    count = 0
    for a1 in range(0, ALT_count + 1):
        for a2 in range(a1, ALT_count + 1):
            ordering[count, 0] = a1
            ordering[count, 1] = a2
            count += 1
        #end for
    #end for

    return ordering
#end def

#################################################################
#    row_gen (original)
#################################################################
def row_gen(GT1, GT2, alt_count, mut_rate):
    ''' '''
    N = alt_count
    combos = (N + 1) * (N + 2) // 2
    row = np.zeros(combos)
    count = 0
    for a1 in range(N + 1):
        for a2 in range(N + 1):
            for a3 in range(N + 1):
                for a4 in range(N + 1):
                    P = 1.0
                    if a1 == GT1[0]:
                        P = P * (1 - mut_rate)
                    else:
                        P = P * mut_rate / N
                    #end if
                    if a2 == GT1[1]:
                        P = P * (1 - mut_rate)
                    else:
                        P = P * mut_rate / N
                    #end if
                    if a3 == GT2[0]:
                        P = P * (1 - mut_rate)
                    else:
                        P = P * mut_rate / N
                    #end if
                    if a4 == GT2[1]:
                        P = P * (1 - mut_rate)
                    else:
                        P = P * mut_rate / N
                    #end if
                    count += 1

                    for b1 in [a1, a2]:
                        for b2 in [a3, a4]:
                            gt_work = np.sort([b1, b2])
                            index = (2 * N + 3 - gt_work[0]) * gt_work[0] // 2 + gt_work[1] - gt_work[0]
                            row[index] = row[index] + 0.25 * P
                        #end for
                    #end for
                #end for
            #end for
        #end for
    #end for

    return row
#end def

#################################################################
#    table_gen (original)
#################################################################
def table_gen(alt_count, mut_rate):
    ''' '''
    N = alt_count
    II_prev = -1

    combos = (N + 1) * (N + 2) // 2
    table = np.zeros([combos ** 2, combos])
    for a1 in range(N + 1):
        for a2 in range(a1, N + 1):
            for a3 in range(N + 1):
                for a4 in range(a3, N + 1):
                    GT1 = [a1, a2]
                    GT2 = [a3, a4]
                    I1 = (2 * N + 3 - GT1[0]) * GT1[0] // 2 + GT1[1] - GT1[0]
                    I2 = (2 * N + 3 - GT2[0]) * GT2[0] // 2 + GT2[1] - GT2[0]
                    II = I1 * combos + I2

                    if II <= II_prev:
                        sys.exit("\nERROR in II calc!!!\n")
                    #end if

                    row = row_gen(GT1, GT2, alt_count, mut_rate)
                    table[II, :] = row
                #end for
            #end for
        #end for
    #end for

    return table
#end def

#################################################################
#    GT_likelihood_wrt_allele_calc (original)
#################################################################
def GT_likelihood_wrt_allele_calc(ALT_count):
    ''' '''
    ordering = GT_ordering_alternate(ALT_count)

    combos = (ALT_count + 1) * (ALT_count + 2) // 2
    table = np.zeros([combos, ALT_count + 1]) * 1.

    for i in range(combos):
        a1 = int(ordering[i, 0]) # casting to int to be used as indexes
        a2 = int(ordering[i, 1])
        table[i, a1] = table[i, a1] + 0.5
        table[i, a2] = table[i, a2] + 0.5
    #end for

    return table
#end def

#################################################################
#    buffering_bams
#################################################################
def buffering_bams(bams_infofile):
    ''' return a list containing reading buffers to bam files (pysam),
    return also a list with the corrisponding IDs associated
    to the bam files '''
    bamfiles, IDs  = [], []
    with open(bams_infofile) as fi:
        for line in fi:
            line_strip = line.rstrip()
            if line_strip:
                try:
                    ID, filepath = line_strip.split('\t') #ID    path/to/file
                    bamfile = pysam.AlignmentFile(filepath, "rb" )
                    IDs.append(ID)
                    bamfiles.append(bamfile)
                except Exception:
                    sys.exit('\nERROR in parsing BAMs info file\n')
                #end try
            #end if
        #end for
    #end with

    return bamfiles, IDs
#end def

#################################################################
#    buffering_rcks
#################################################################
def buffering_rcks(rcks_infofile):
    ''' return a list containing reading buffers to rck files (tabix),
    return also a list with the corrisponding IDs associated to files'''
    rckfiles, IDs  = [], []
    with open(rcks_infofile) as fi:
        for line in fi:
            line_strip = line.rstrip()
            if line_strip:
                try:
                    ID, filepath = line_strip.split('\t') #ID    path/to/file
                    rckfile = tabix.open(filepath)
                    IDs.append(ID)
                    rckfiles.append(rckfile)
                except Exception:
                    sys.exit('\nERROR in parsing RCKs info file\n')
                #end try
            #end if
        #end for
    #end with

    return rckfiles, IDs
#end def

#################################################################
#    get_ADs_bam (original refactored)
#################################################################
def get_ADs_bam(bamfile, chrom, pos, REF, MQthr, BQthr, deletion=False, insertion=False):
    ''' access bam file and return AD counts by strand for reference and alternate allele for variant,
    the way pileup is used depend on variant type (insertion, delition or snv) '''
    # coordinates in pysam are always 0-based
    #   -> current position - 1 to access the right position from pysam
    position = pos - 1
    REF_upper = REF.upper()
    ADf, ADr = np.array([0., 0]), np.array([0., 0])

    # Getting pileup info
    try: SP = bamfile.pileup(chrom, position, position + 1)
    except Exception:
        sys.exit('\nERROR in accessing BAM file: variant and file chromosome formats are not matching\n')
    #end try

    # Getting AD info
    for pileupcolumn in SP:
        if pileupcolumn.pos == position:
            for pileupread in pileupcolumn.pileups:
                if not pileupread.is_del and not pileupread.is_refskip:
                    qp = pileupread.query_position
                    MQ = pileupread.alignment.mapping_quality
                    BQ = pileupread.alignment.query_qualities[qp]
                    if MQ >= MQthr and BQ >= BQthr:
                        seq = pileupread.alignment.query_sequence[qp].upper()
                        # DELETION
                        if deletion:
                            indel_val = pileupread.indel
                            if seq == REF_upper and indel_val >= 0:
                                if pileupread.alignment.is_reverse: ADr[0] += 1
                                else: ADf[0] += 1
                                #end if
                            elif seq == REF_upper and indel_val < 0:
                                if pileupread.alignment.is_reverse: ADr[1] += 1
                                else: ADf[1] += 1
                                #end if
                            #end if
                        # INSERTION
                        elif insertion:
                            indel_val = pileupread.indel
                            if seq == REF_upper and indel_val <= 0:
                                if pileupread.alignment.is_reverse: ADr[0] += 1
                                else: ADf[0] += 1
                                #end if
                            elif seq == REF_upper and indel_val > 0:
                                if pileupread.alignment.is_reverse: ADr[1] += 1
                                else: ADf[1] += 1
                                #end if
                            #end if
                        # SNV
                        else:
                            if seq == REF_upper:
                                if pileupread.alignment.is_reverse: ADr[0] += 1
                                else: ADf[0] += 1
                                #end if
                            else:
                                if pileupread.alignment.is_reverse: ADr[1] += 1
                                else: ADf[1] += 1
                                #end if
                            #end if
                        #end if
                    #end if
                #end if
            #end for
        #end if
    #end for

    return ADf, ADr
#end def

#################################################################
#    get_ADs_rck
#################################################################
def get_ADs_rck(rckfile, chrom, pos, deletion=False, insertion=False):
    ''' '''
    # rck format is 1-based following mpileup standard (1-based position on the chromosome)
    region = chrom + ':' + str(pos) + '-' + str(pos)
    chr, rec_pos, cov, ref_fw, ref_rv, alt_fw, alt_rv, \
        ins_fw, ins_rv, del_fw, del_rv = next(rckfile.querys(region))
    if pos != int(rec_pos):
        raise IndexError('\nERROR in RCK file indexing: position received is not consistent with position called\n')
    #end if
    if deletion:
        ADf = np.array([int(ref_fw), int(del_fw)])
        ADr = np.array([int(ref_rv), int(del_rv)])
    elif insertion:
        ADf = np.array([int(ref_fw), int(ins_fw)])
        ADr = np.array([int(ref_rv), int(ins_rv)])
    else:
        ADf = np.array([int(ref_fw), int(alt_fw)])
        ADr = np.array([int(ref_rv), int(alt_rv)])
    #end if

    return ADf, ADr
#end def

#################################################################
#    get_ADs_caller
#################################################################
def get_ADs_caller(file, chrom, pos, REF, ALT, MQthr, BQthr, is_bam):
    ''' define variant type and run the appropriate function to access file
    to retrieve AD counts by strand for reference and alternate allele for variant'''
    split_ALT = ALT.split(',')
    if len(split_ALT) > 1:
        if is_bam:
            return get_ADs_bam(file, chrom, pos, REF[0], MQthr, BQthr)
        else:
            return get_ADs_rck(file, chrom, pos)
        #end if
    elif len(REF) > 1:
        if is_bam:
            return get_ADs_bam(file, chrom, pos, REF[0], MQthr, BQthr, deletion=True)
        else:
            return get_ADs_rck(file, chrom, pos, deletion=True)
        #end if
    elif len(ALT) > 1:
        if is_bam:
            return get_ADs_bam(file, chrom, pos, REF[0], MQthr, BQthr, insertion=True)
        else:
            return get_ADs_rck(file, chrom, pos, insertion=True)
        #end if
    #end if

    if is_bam:
        return get_ADs_bam(file, chrom, pos, REF[0], MQthr, BQthr)
    else:
        return get_ADs_rck(file, chrom, pos)
    #end if
#end def

#################################################################
#    get_all_ADs
#################################################################
def get_all_ADs(files, chrom, pos, REF, ALT, MQthr, BQthr, is_bam):
    ''' return the AD counts by strand for reference and alternate allele for variant
    in all files '''
    ADfs, ADrs = [], []
    for file in files:
        ADf, ADr = get_ADs_caller(file, chrom, pos, REF, ALT, MQthr, BQthr, is_bam)
        ADfs.append(ADf)
        ADrs.append(ADr)
    #end for

    return np.array(ADfs), np.array(ADrs)
#end def


#################################################################
#    M1_L_calc_aux (original)
#################################################################
def M1_L_calc_aux(rho, k):
    ''' '''
    ALT_count = 1
    M1_L_k = np.zeros(ALT_count + 1)
    default = np.log((1. - rho) / ALT_count)
    M1_L_k = M1_L_k + default
    M1_L_k[k] = np.log(rho)

    return M1_L_k
#end def

#################################################################
#    M1_L_calc (original)
#################################################################
def M1_L_calc(AD, rho):
    ''' '''
    ALT_count = 1
    if AD.size - 1 != ALT_count:
        sys.exit("\nERROR in M1_L_calc\n")
    #end if
    M1_L = []
    for k in range(ALT_count + 1):
        M1_L_k = M1_L_calc_aux(rho,k)
        M1_L.append(M1_L_k)
    #end for

    return M1_L
#end def

#################################################################
#    M2_L_calc_aux (original)
#################################################################
def M2_L_calc_aux(M1_L_k, GT_likelihood_wrt_allele_L):
    ''' '''
    ALT_count = 1
    if M1_L_k.size - 1 != ALT_count:
        sys.exit("\nERROR in M2_L_calc_aux\n")
    #end if
    combos = (ALT_count + 1) * (ALT_count + 2) // 2
    temp_table = GT_likelihood_wrt_allele_L + np.tile(M1_L_k.reshape([1, ALT_count + 1]),[combos, 1])
    M2_L_k = np.zeros(combos)
    for i in range(combos):
        row = temp_table[i, :]
        row_max = np.max(row)
        row = row - row_max
        M2_L_k[i] = np.log(np.sum(np.exp(row))) + row_max
    #end for

    return M2_L_k
#end def

#################################################################
#    M2_L_calc (original)
#################################################################
def M2_L_calc(M1_L, GT_likelihood_wrt_allele_L):
    ''' '''
    ALT_count = 1
    if (M1_L[0]).size - 1 != ALT_count:
        sys.exit("\nERROR in M2_L_calc\n")
    #end if
    M2_L = []
    for k in range(ALT_count + 1):
        M1_L_k = M1_L[k]
        M2_L_k = M2_L_calc_aux(M1_L_k,GT_likelihood_wrt_allele_L)
        M2_L.append(M2_L_k)
    #end for

    return M2_L
#end def

#################################################################
#    GT_marg_L_calc (original)
#################################################################
def GT_marg_L_calc(M2_L_f, M2_L_r, ADf, ADr, prior_L):
    ''' '''
    GT_marg_L = prior_L
    ALT_count = 1
    if ADf.size - 1 != ALT_count or ADr.size - 1 != ALT_count:
        sys.exit("\nERROR in GT_marg_L_calc\n")
    #end if
    for k in range(ALT_count + 1):
        M2_L_k = M2_L_f[k]
        GT_marg_L = GT_marg_L + ADf[k] * M2_L_k
    #end for
    for k in range(ALT_count + 1):
        M2_L_k = M2_L_r[k]
        GT_marg_L = GT_marg_L + ADr[k] * M2_L_k
    #end for

    return GT_marg_L
#end def

#################################################################
#    M3_L_calc_aux (original)
#################################################################
def M3_L_calc_aux(GT_marg_L, M2_L_k):
    ''' '''
    M3_L_k = GT_marg_L - M2_L_k

    return M3_L_k
#end def

#################################################################
#    M3_L_calc (original)
#################################################################
def M3_L_calc(GT_marg_L, M2_L):
    ''' '''
    ALT_count = 1
    if len(M2_L) - 1 != ALT_count:
        sys.exit("\nERROR in M3_L_calc\n")
    #end if
    M3_L = []
    for k in range(ALT_count + 1):
        M2_L_k = M2_L[k]
        M3_L_k = M3_L_calc_aux(GT_marg_L, M2_L_k)
        M3_L.append(M3_L_k)
    #end for

    return M3_L
#end def

#################################################################
#    M4_L_calc_aux (original)
#################################################################
def M4_L_calc_aux(M3_L_k, GT_likelihood_wrt_allele_L):
    ''' '''
    ALT_count = 1
    if (GT_likelihood_wrt_allele_L.shape)[1] - 1 != 1:
        sys.exit("\nERROR in M4_L_calc_aux\n")
    #end if
    combos = (ALT_count + 1) * (ALT_count + 2) // 2
    temp_table = GT_likelihood_wrt_allele_L + np.tile(M3_L_k.reshape([combos, 1]), [1, ALT_count + 1])
    M4_L_k = np.zeros(ALT_count + 1)
    for i in range(ALT_count + 1):
        column = temp_table[:, i]
        column_max = np.max(column)
        column = column - column_max
        M4_L_k[i] = np.log(np.sum(np.exp(column))) + column_max
    #end for

    return M4_L_k
#end def

#################################################################
#    M4_L_calc (original)
#################################################################
def M4_L_calc(M3_L, GT_likelihood_wrt_allele_L):
    ''' '''
    ALT_count = 1
    if (GT_likelihood_wrt_allele_L.shape)[1] - 1 != ALT_count:
        sys.exit("\nERROR in M4_L_calc\n")
    #end if
    M4_L = []
    for k in range(ALT_count + 1):
        M3_L_k = M3_L[k]
        M4_L_k = M4_L_calc_aux(M3_L_k, GT_likelihood_wrt_allele_L)
        M4_L.append(M4_L_k)
    #end for

    return M4_L
#end def

#################################################################
#    A_marg_L_calc (original)
#################################################################
def A_marg_L_calc(M1_L, M4_L):
    ''' '''
    ALT_count = 1
    if len(M1_L) - 1 != ALT_count:
        sys.exit("\nERROR in A_marg_L_calc\n")
    #end if
    A_marg_L = []
    for k in range(ALT_count + 1):
        M1_L_k = M1_L[k]
        M4_L_k = M4_L[k]
        A_marg_L_k = M1_L_k + M4_L_k
        A_marg_L.append(A_marg_L_k)
    #end for

    return A_marg_L
#end def

#################################################################
#    T_term_calc_for_rho (original)
#################################################################
def T_term_calc_for_rho(A_marg_L, AD):
    ''' '''
    if len(A_marg_L) != AD.size:
        sys.exit("\nERROR in T_term_calc\n")
    #end if
    ALT_count = AD.size - 1
    T1_term = 0.
    T2_term = 0.
    for k in range(ALT_count + 1):
        A_marg_L_k = A_marg_L[k]
        A_marg_temp = np.exp(A_marg_L_k - np.max(A_marg_L_k))
        A_marg = A_marg_temp / np.sum(A_marg_temp)
        T1_term = T1_term + A_marg[k] * AD[k]
        T2_term = T2_term + (1. - A_marg[k]) * AD[k]
    #end for

    return T1_term, T2_term
#end def

#################################################################
#    GT_marg_L_to_GT_marg (original)
#################################################################
def GT_marg_L_to_GT_marg(GT_marg_L):
    ''' '''
    M = np.max(GT_marg_L)
    GT_marg_L = GT_marg_L - M
    GT_marg = np.exp(GT_marg_L)
    S = np.sum(GT_marg)
    GT_marg = GT_marg / S
    joint_probty_term = np.log(S) + M

    return GT_marg, joint_probty_term
#end def

#################################################################
#    EM_step (original)
#################################################################
def EM_step(ADf_list, ADr_list, rho_f_old, rho_r_old, prior_L_old, GT_likelihood_wrt_allele_L, a, b, D_original, allele_freq):
    ''' '''
    D = np.zeros(3)
    D[0] = D_original[0]
    D[1] = D_original[1]
    D[2] = D_original[2]

    if allele_freq <= 0.:
        AF = 0.
    else:
        AF = allele_freq
    #end if

    f0 = (1. - AF) ** 2.
    f2 = AF ** 2.
    f1 = 1. - f0 - f2
    D = np.array([f0, f1, f2]) * 1000. + 2.

    T1_f = a - 1.
    T2_f = b - 1.
    T1_r = a - 1.
    T2_r = b - 1.
    T_for_prior = D - 1.
    joint_probty =                (a - 1.) * np.log(rho_f_old) + (b - 1.) * np.log(1. - rho_f_old)
    joint_probty = joint_probty + (a - 1.) * np.log(rho_r_old) + (b - 1.) * np.log(1. - rho_r_old)
    for i in range(3):
        joint_probty = joint_probty + (D[i] - 1) * prior_L_old[i]
    #end for

    if len(ADf_list) != len(ADr_list):
        sys.exit("\nERROR1 in EM_step\n")
    #end if

    for i in range(len(ADf_list)):
        ADf = ADf_list[i]
        ADr = ADr_list[i]
        M1_L_f = M1_L_calc(ADf, rho_f_old)
        M1_L_r = M1_L_calc(ADr, rho_r_old)
        M2_L_f = M2_L_calc(M1_L_f, GT_likelihood_wrt_allele_L)
        M2_L_r = M2_L_calc(M1_L_r, GT_likelihood_wrt_allele_L)
        GT_marg_L = GT_marg_L_calc(M2_L_f, M2_L_r, ADf, ADr, prior_L_old)
        M3_L_f = M3_L_calc(GT_marg_L, M2_L_f)
        M3_L_r = M3_L_calc(GT_marg_L, M2_L_r)
        M4_L_f = M4_L_calc(M3_L_f, GT_likelihood_wrt_allele_L)
        M4_L_r = M4_L_calc(M3_L_r, GT_likelihood_wrt_allele_L)
        A_marg_L_f = A_marg_L_calc(M1_L_f, M4_L_f)
        A_marg_L_r = A_marg_L_calc(M1_L_r, M4_L_r)

        T1_term_f, T2_term_f = T_term_calc_for_rho(A_marg_L_f, ADf)
        T1_term_r, T2_term_r = T_term_calc_for_rho(A_marg_L_r, ADr)

        T1_f = T1_f + T1_term_f
        T2_f = T2_f + T2_term_f
        T1_r = T1_r + T1_term_r
        T2_r = T2_r + T2_term_r

        GT_marg,joint_probty_term = GT_marg_L_to_GT_marg(GT_marg_L)
        joint_probty = joint_probty + joint_probty_term

        T_for_prior = T_for_prior + GT_marg
    #end for

    rho_f_new = 1. / (1. + T2_f / T1_f)
    rho_r_new = 1. / (1. + T2_r / T1_r)
    prior_new = T_for_prior / np.sum(T_for_prior)
    prior_L_new = np.log(prior_new)

    return rho_f_new, rho_r_new, prior_L_new, joint_probty
#end def

#################################################################
#    EM_full (original)
#################################################################
def EM_full(ADfs, ADrs, rho_f_old, rho_r_old, prior_L_old, GT_likelihood_wrt_allele_L, a, b, D, allele_freq):
    ''' '''
    joint_probty_s = []
    joint_probty_new = np.nan
    for i in range(3):
        joint_probty_old = joint_probty_new
        rho_f_new, rho_r_new, prior_L_new, joint_probty_new = \
            EM_step(ADfs, ADrs, rho_f_old, rho_r_old, prior_L_old, GT_likelihood_wrt_allele_L, a, b, D, allele_freq)
        rho_f_old = rho_f_new
        rho_r_old = rho_r_new
        prior_L_old = prior_L_new
        joint_probty_s.append(joint_probty_new)
    #end for

    while np.abs(joint_probty_old - joint_probty_new) > 10 ** -7:
        joint_probty_old = joint_probty_new
        rho_f_new, rho_r_new, prior_L_new, joint_probty_new = \
            EM_step(ADfs, ADrs, rho_f_old, rho_r_old, prior_L_old, GT_likelihood_wrt_allele_L, a, b, D, allele_freq)
        rho_f_old = rho_f_new
        rho_r_old = rho_r_new
        prior_L_old = prior_L_new
        joint_probty_s.append(joint_probty_new)
    #end while

    return rho_f_new, rho_r_new, prior_L_new, joint_probty_s
#end def

#################################################################
#    GTL_L_calc (original)
#################################################################
def GTL_L_calc(ADf, ADr, rho_f, rho_r, GT_likelihood_wrt_allele_L):
    ''' '''
    M1_L_f = M1_L_calc(ADf, rho_f)
    M1_L_r = M1_L_calc(ADr, rho_r)
    M2_L_f = M2_L_calc(M1_L_f, GT_likelihood_wrt_allele_L)
    M2_L_r = M2_L_calc(M1_L_r, GT_likelihood_wrt_allele_L)
    prior_L = np.zeros(3)
    GTL_L = GT_marg_L_calc(M2_L_f, M2_L_r, ADf, ADr, prior_L)
    GTL_L = GTL_L - np.max(GTL_L)

    return GTL_L
#end def

#################################################################
#    posterior_probty_calc_exact (original)
#################################################################
def posterior_probty_calc_exact(prior_L, table_L, C_GL_L, M_GL_L, D_GL_L):
    ''' '''
    combos = 3
    work_column = np.empty(combos ** 2)
    for I1 in range(combos):
        for I2 in range(combos):
            II = I1 * combos + I2
            work_column[II] = prior_L[I1] + prior_L[I2] + M_GL_L[I1] + D_GL_L[I2]
        #end for
    #end for

    work_table = table_L + np.tile(C_GL_L, [combos ** 2, 1]) + np.tile(np.reshape(work_column, [combos ** 2, 1]), [1, combos])
    work_table = work_table - np.max(work_table)
    work_table = np.exp(work_table)
    work_table = work_table / np.sum(work_table)
    PP = np.max(np.array([work_table[0][1], work_table[0][2]]))

    return PP, work_table
#end def

#################################################################
#    denovo_P_calc (original)
#################################################################
def denovo_P_calc(ADfs, ADrs, rho_f, rho_r, GT_likelihood_wrt_allele_L, table_L, prior_L):
    ''' '''
    M_GL_L = GTL_L_calc(ADfs[0], ADrs[0], rho_f, rho_r, GT_likelihood_wrt_allele_L)
    D_GL_L = GTL_L_calc(ADfs[1], ADrs[1], rho_f, rho_r, GT_likelihood_wrt_allele_L)
    C_GL_L = GTL_L_calc(ADfs[2], ADrs[2], rho_f, rho_r, GT_likelihood_wrt_allele_L) # child is the last one
    PP, work_table = posterior_probty_calc_exact(prior_L, table_L, C_GL_L, M_GL_L, D_GL_L)

    return PP, work_table
#end def

#################################################################
#    PP_calc (original)
#################################################################
def PP_calc(trio_files, unrelated_files, chrom, pos, REF, ALT, allele_freq, MQthresh, BQthresh, is_bam):
    ''' '''
    ADfs_U, ADrs_U = get_all_ADs(unrelated_files, chrom, pos, REF, ALT, MQthresh, BQthresh, is_bam)
    rho_f_old, rho_r_old = 0.8, 0.8
    prior_old = np.array([1. / 3, 1. / 3, 1. / 3])
    prior_old = prior_old / np.sum(prior_old)
    prior_L_old = np.log(prior_old)
    GT_likelihood_wrt_allele = GT_likelihood_wrt_allele_calc(1)
    GT_likelihood_wrt_allele_L = np.log(GT_likelihood_wrt_allele)
    a, b, D = 2., 2., np.array([2., 2, 2])

    rho_f_new, rho_r_new, prior_L_new, joint_probty_s = \
        EM_full(ADfs_U, ADrs_U, rho_f_old, rho_r_old, prior_L_old, GT_likelihood_wrt_allele_L, a, b, D, allele_freq)

    AF_unrel = 0.
    for i in range(ADfs_U.shape[0]):
        temp1 = GTL_L_calc(ADfs_U[i], ADrs_U[i], rho_f_new, rho_r_new, GT_likelihood_wrt_allele_L)
        temp = temp1 + prior_L_new
        temp = temp - np.max(temp)
        temp = np.exp(temp)
        temp = temp / np.sum(temp)

        AF_unrel = AF_unrel + temp[1] + temp[2] * 2.
    #end for

    AF_unrel = AF_unrel / 2. / ADfs_U.shape[0]

    ADfs, ADrs = get_all_ADs(trio_files, chrom, pos, REF, ALT, MQthresh, BQthresh, is_bam)

    table = table_gen(1, 1e-8)
    table_L = np.log(table)
    PP, work_table = denovo_P_calc(ADfs, ADrs, rho_f_new, rho_r_new, GT_likelihood_wrt_allele_L, table_L, prior_L_new)

    return PP, ADfs, ADrs, ADfs_U, ADrs_U, rho_f_new, rho_r_new, prior_L_new, AF_unrel
#end def

#################################################################
#    ALT_count_check_parents
#################################################################
def ALT_count_check_parents(ADfs, ADrs, thr):
    ''' check if total alternate reads count in parents is over threshold '''
    if len(ADfs) != 3 or len(ADrs) != 3:
        sys.exit("\nERROR in retrieving stranded AD counts: missing information for trio\n")
    #end if
    alt_count = ADfs[0][1] + ADfs[1][1] + ADrs[0][1] + ADrs[1][1]

    if alt_count > thr: return True
    else: return False
    #end if
#end def

#################################################################
#    runner
#################################################################
def main(args, test=False):
    ''' '''
    # Variables
    is_bam = True if args['bam'] else False
    is_afthr = True if args['afthr'] else False
    afthr, aftag, aftag_idx = 1., 'novoAF', 0 # novoAF as aftag placeholder if not is_afthr
    ppthr = float(args['ppthr']) if args['ppthr'] else 0.
    afthr_unrelated = float(args['afthr_unrelated']) if args['afthr_unrelated'] else 1.
    MQthr = int(args['MQthr']) if args['MQthr'] else 0
    BQthr = int(args['BQthr']) if args['BQthr'] else 0
    ADthr = int(args['ADthr']) if args['ADthr'] else 0
    RSTR_def = '##FORMAT=<ID=RSTR,Number=4,Type=Integer,Description="Read counts by strand for ref and alt alleles (Rf,Af,Rr,Ar)">'
    novoCaller_def = '##INFO=<ID=novoPP,Number=1,Type=Float,Description="Posterior probability from novoCaller">'
    # NA chromosomes set -> import from shared_vars
    if test: NA_chroms = test_NA_chroms
    else: NA_chroms = real_NA_chroms
    #end if
    is_NA = False
    is_verbose = True if args['verbose'] else False

    # Buffers
    fo = open(args['outputfile'], 'w')

    # Creating Vcf object
    vcf_obj = vcf_parser.Vcf(args['inputfile'])

    # Check arguments
    if is_afthr:
        afthr = float(args['afthr'])
        if args['aftag']:
            aftag, aftag_idx = vcf_obj.header.check_tag_definition(args['aftag'])
        else:
            sys.exit('\nERROR in parsing arguments: to filter by population allele frequency please specify the TAG to use\n')
        #end if
    #end if

    # Data structures
    variants_passed = []

    # Getting files and associated IDs
    sys.stderr.write('Getting unrelated and trio files...\n')
    sys.stderr.flush()

    if is_bam: # if bam files
        unrelated_files, IDs_unrelated = buffering_bams(args['unrelatedfiles'])
        trio_files, IDs_trio = buffering_bams(args['triofiles']) # [parent, parent, child]
    else:
        unrelated_files, IDs_unrelated = buffering_rcks(args['unrelatedfiles'])
        trio_files, IDs_trio = buffering_rcks(args['triofiles']) # [parent, parent, child]
    #end if

    # Checking info files for trio is complete
    if len(trio_files) != 3:
        sys.exit('\nERROR in BAMs info file for trio: missing information for some family member\n')
    #end if

    # Checking information for trio is complete in the vcf
    for ID in IDs_trio:
        if ID not in vcf_obj.header.IDs_genotypes:
            sys.exit('\nERROR in VCF file: missing information for some family member\n')
        #end if
    #end for

    # Reading variants
    analyzed = 0
    for i, vnt_obj in enumerate(vcf_obj.parse_variants()):
        if is_verbose:
            sys.stderr.write('\rAnalyzing variant... ' + str(i + 1))
            sys.stderr.flush()
        #end if

        # # Check if chromosome is canonical and in valid format
        # if not check_chrom(vnt_obj.CHROM): # skip variant if not
        #     continue
        # #end if

        # Getting allele frequency from novoAF tag
        af = allele_frequency(vnt_obj, aftag, aftag_idx)

        # is_NA reset
        is_NA = False

        # Calculate statistics
        if af <= afthr: # hard filter on allele frequency
            analyzed += 1
            PP, ADfs, ADrs, ADfs_U, ADrs_U, _, _, _, AF_unrel = \
                PP_calc(trio_files, unrelated_files, vnt_obj.CHROM, vnt_obj.POS, vnt_obj.REF, vnt_obj.ALT, af, MQthr, BQthr, is_bam)
            if vnt_obj.CHROM.replace('chr', '') in NA_chroms:
            # model assumptions does not apply to sex and mithocondrial chromosomes, PP -> NA
                PP = 0.
                is_NA = True
            #end if
            if ADthr and ALT_count_check_parents(ADfs, ADrs, ADthr):
            # AD in parents over ADthr, PP -> 0
                PP = 0.
                is_NA = False
            #end if
            if AF_unrel <= afthr_unrelated and PP >= ppthr: # hard filter on AF_unrel, PP
                variants_passed.append([PP, ADfs, ADrs, ADfs_U, ADrs_U, AF_unrel, is_NA, vnt_obj])
            #end if
        #end if
    #end for

    # Writing output
    sys.stderr.write('\n\n...Writing results for ' + str(analyzed) + ' analyzed variants out of ' + str(i + 1) + ' total variants\n')
    sys.stderr.flush()

    # Header definitions
    is_RSTR = 'RSTR' in vcf_obj.header.definitions
    is_novoCaller = 'novoPP' in vcf_obj.header.definitions

    if not is_RSTR:
        vcf_obj.header.add_tag_definition(RSTR_def, 'FORMAT')
    #end if
    if not is_novoCaller:
        vcf_obj.header.add_tag_definition(novoCaller_def, 'INFO')
    #end if
    vcf_obj.write_definitions(fo)

    # Adding to header columns unrelated samples missing IDs
    fo.write(vcf_obj.header.columns.rstrip())
    for ID in IDs_unrelated:
        if ID not in vcf_obj.header.IDs_genotypes:
            fo.write('\t' + ID)
        #end if
    #end for
    fo.write('\n')

    # Variants passed
    for variant in sorted(variants_passed, key=lambda x: x[0], reverse=True):
        PP, ADfs, ADrs, ADfs_U, ADrs_U, AF_unrel, is_NA, vnt_obj = variant

        # Removing older tags fields if present
        if is_RSTR:
            vnt_obj.remove_tag_genotype('RSTR')
        #end if
        if is_novoCaller:
            vnt_obj.remove_tag_info('novoPP')
        #end if

        # Adding new tag
        if not is_NA:
            vnt_obj.add_tag_info('novoPP={0}'.format(PP))
        #end if

        # Fill the trailing fields dropped in genotypes
        vnt_obj.complete_genotype()

        # Updating genotypes trio
        for i, ID in enumerate(IDs_trio):
            values = '{0},{1},{2},{3}'.format(int(ADfs[i][0]), int(ADfs[i][1]), int(ADrs[i][0]), int(ADrs[i][1]))
            vnt_obj.add_values_genotype(ID, values)
        #end for

        # Updating genotypes unrelated
        unrelated_genotypes = []
        for i, ID in enumerate(IDs_unrelated):
            values = '{0},{1},{2},{3}'.format(int(ADfs_U[i][0]), int(ADfs_U[i][1]), int(ADrs_U[i][0]), int(ADrs_U[i][1]))
            if ID in vnt_obj.GENOTYPES:
                vnt_obj.add_values_genotype(ID, values)
            else:
                unrelated_genotypes.append(vnt_obj.empty_genotype() + ':' + values)
            #end if
        #end for

        # Updating FORMAT
        vnt_obj.add_tag_format('RSTR')

        # Writing output
        if unrelated_genotypes:
            fo.write(vnt_obj.to_string().rstrip() + '\t' + '\t'.join(unrelated_genotypes) + '\n')
        else:
            vcf_obj.write_variant(fo, vnt_obj)
        #end if
    #end for

    # Closing files buffers
    fo.close()
    if is_bam:
        for buffer in unrelated_files:
            buffer.close()
        #end for
        for buffer in trio_files:
            buffer.close()
        #end for
    #end if
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python

#################################################################
#
#    mpileupCounts
#        note: require samtools available in environment
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import subprocess
# mpileup_parser
from granite.lib import mpileup_parser
# fasta_parser
from granite.lib import fasta_parser



#################################################################
#
#    FUNCTIONS
#
#################################################################
def run_mpileupParser(fi, fo, ref_dict):
    ''' '''
    mP = mpileup_parser.mpileupParser()
    first = True
    for mC in mP.generator(fi):
        try: mC.get_AD_noreference(ref_dict[mC.chr][mC.pos-1])
        except Exception:
            sys.exit('\nERROR in reading position information: chr format ({0}) is not matching reference\n'
                      .format(mC.chr))
        #end try
        if first:
            first = False
            mC.write_AD(fo, header=True)
        else:
            mC.write_AD(fo)
        #end if
    #end for
#end def

#################################################################
#    runner
#################################################################
def main(args):
    ''' '''
    # Initialize objects and variables
    handler = fasta_parser.FastaHandler()
    ref_dict = {} # {chr: seq, ...}
    is_chr, chr = False, ''

    # Parsing region if available
    if args['region']:
        is_chr = True
        if ':' in args['region']:
            try:
                chr, region = args['region'].split(':')
                strt, end = map(int, region.split('-'))
                if strt >= end:
                    sys.exit('\nERROR in parsing region argument: start index is larger than end index\n')
                #end if
            except Exception:
                sys.exit('\nERROR in parsing region argument: the format is not recognized\n')
            #end try
        else:
            try:
                chr = args['region']
            except Exception:
                sys.exit('\nERROR in parsing region argument: the format is not recognized\n')
            #end try
        #end if
    #end if

    # Building command line
    command_line = ['samtools', 'mpileup', '-aa']
    if is_chr: command_line += ['-r', args['region']]
    #end if
    if args['MQthr']: command_line += ['--min-MQ', str(args['MQthr'])]
    #end if
    if args['BQthr']: command_line += ['--min-BQ', str(args['BQthr'])]
    #end if
    command_line += [args['inputfile']]

    # Reading reference into iterator
    IT = handler.parse_generator(args['reference'])

    # Output
    fo = open(args['outputfile'], 'w')

    # Loading reference
    if is_chr:
        for header, seq in IT:
            if header.split()[0] == chr:
                ref_dict.setdefault(chr, seq)
                break
            #end if
        #end for
    else:
        for header, seq in IT:
            chr = header.split()[0]
            ref_dict.setdefault(chr, seq)
        #end for
    #end if

    # Running samtools
    pipe_in = subprocess.Popen(command_line, stdout=subprocess.PIPE)

    # Parsing mpileup
    run_mpileupParser(pipe_in.stdout, fo, ref_dict)

    # Closing files
    fo.close()
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python

#################################################################
#
#    pedigree_parser
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os


#################################################################
#
#    Pedigree
#       -> Member
#
#################################################################
class Pedigree(object):
    ''' object to represent a pedigree '''

    class Member(object):
        ''' object to represent a member of the pedigree '''

        def __init__(self, name, sample, gender):
            ''' initialize Member object '''
            self.name = name
            self.sample = sample
            self.gender = gender
            self.parents = []
            self.children = []
        #end def __init__

        def add_parent(self, parent):
            ''' add a Member object that is parent '''
            self.parents.append(parent)
        #end def add_parent

        def add_child(self, child):
            ''' add a Member object that is child '''
            self.children.append(child)
        #end def add_child

        def is_sample(self):
            ''' '''
            if self.sample: return True
            #end if
            return False
        #end def is_sample

        def get_siblings(self):
            ''' return a list of Member objects that are siblings '''
            siblings = []
            for parent in self.parents:
                for child in parent.children:
                    if child not in siblings and child != self:
                        siblings.append(child)
                    #end if
                #end for
            #end for
            return siblings
        #end def get_siblings

        def get_parents(self):
            ''' return a list of Member objects that are parents '''
            return self.parents
        #end def get_parents

        def get_children(self):
            ''' return a list of Member objects that are children '''
            return self.children
        #end def get_children

        def get_spouses(self):
            ''' return a list of Member objects that are spouses,
            sorted by descending number of common children '''
            tmp_spouses = [] # [(num_children, spouse_obj), ...]
            for child in self.children:
                for parent in child.parents:
                    if parent != self:
                        num_children = len(self.common_children(parent))
                        if (num_children, parent) not in tmp_spouses:
                            tmp_spouses.append((num_children, parent))
                        #end if
                    #end if
                #end for
            #end for
            return [tmp_spouse[1] for tmp_spouse in sorted(tmp_spouses, key=lambda x: x[0], reverse=True)]
        #end def get_spouses

        def common_children(self, spouse):
            ''' return a list of Member objects that are children in common with spouse '''
            return list(set(self.children).intersection(set(spouse.children)))
        #end def common_children

        def has_parents(self):
            ''' check if self has two parents with samples '''
            for i, parent in enumerate(self.parents):
                if not parent.sample: # missing parent sample information
                    return False
                #end if
            #end for
            try:
                if i != 1: return False # missing one parent
                #end if
            except Exception: return False # missing both parents
            #end try
            return True
        #end def has_parents

    #end class Member

    def __init__(self, pedigree):
        ''' initialize Pedigree object,
        pedigree information must be provided as json '''
        self.members = {} # dictionary of Member objects by name
        self.samples = {} # dictionary to map sample to name
        self.parse_pedigree(pedigree)
    #end def __init__

    def add_member(self, member):
        ''' create Member object for member,
        member information must be provided as dict '''
        # Get values
        try: name = member['individual'] # individual unique ID, every member must have one
        except Exception:
            raise ValueError('\nERROR in pedigree structure, missing name for pedigree member\n')
        #end try
        try: sample = member['sample_name'] # sample ID, if available
        except Exception: sample = None
        #end try
        gender = member['gender']
        # Create object
        self.members.setdefault(name, self.Member(name, sample, gender))
        if sample: self.samples.setdefault(sample, name)
        #end if
    #end def add_member

    def add_parent(self, name, parent_name):
        ''' add parent-child relation to parent and child by name '''
        if name not in self.members:
            raise ValueError('\nERROR in pedigree structure, missing name {0} in pedigree\n'
                             .format(name))
        #end if
        if parent_name not in self.members:
            return
        #end if
        self.members[name].add_parent(self.members[parent_name])
        self.members[parent_name].add_child(self.members[name])
    #end def add_parent

    def get_member_by_sample(self, sample):
        ''' return Member object by sample '''
        try:
            return self.members[self.samples[sample]]
        except Exception:
            raise ValueError('\nERROR in pedigree structure, missing sample {0} in pedigree\n'
                             .format(sample))
        #end try
    #end def get_member_by_sample

    def get_family(self, sample):
        ''' given sample in pedigree,
        build family structure around it '''
        family = {
            'children': [],
            'parents': []
        }
        sample_obj = self.get_member_by_sample(sample)
        if sample_obj.get_children(): # sample is the center of the family,
                                      # check for parents and children
            if sample_obj.has_parents(): # parents information complete
                spouse = None
                for spouse_ in sample_obj.get_spouses():
                    if spouse_.has_parents():
                        spouse = spouse_
                        break
                    #end if
                #end for
                if not spouse:
                    raise ValueError('\nERROR in building family from pedigree, missing parents information for sample {0} spouse\n'
                            .format(sample))
                #end if
                # Create family
                family['children'] = sample_obj.common_children(spouse)
                family['parents'] = [obj for obj in sorted([sample_obj, spouse], key=lambda x: x.name)]
            else: # missing parents
                raise ValueError('\nERROR in building family from pedigree, missing parents information for sample {0}\n'
                        .format(sample))
            #end if
        else: # sample is the newest generation,
              # check for siblings, parents and grandparents
            if sample_obj.has_parents(): # parents information complete, check grandparents
                parents = sample_obj.get_parents()
                for parent in parents:
                    if not parent.has_parents():
                        raise ValueError('\nERROR in building family from pedigree, missing family information for sample {0}\n'
                                .format(sample))
                    #end if
                #end for
                # Create family
                family['children'] = parents[0].common_children(parents[1])
                family['parents'] = [obj for obj in sorted(parents, key=lambda x: x.name)]
            else:
                raise ValueError('\nERROR in building family from pedigree, missing parents information for sample {0}\n'
                        .format(sample))
            #end if
        #end if
        return family
    #end def get_family

    def parse_pedigree(self, pedigree):
        ''' read pedigree information to build Pedigree object,
        pedigree information must be provided as json '''
        # Creating Member objects
        for member in pedigree:
            self.add_member(member)
        #end for
        # Adding relations
        for member in pedigree:
            name = member['individual']
            if len(member['parents']) > 2:
                raise ValueError('\nERROR in pedigree structure, {0} has more than two parents\n'
                                 .format(name))
            #end if
            for parent_name in member['parents']:
                self.add_parent(name, parent_name)
            #end for
        #end for
    #end def parse_pedigree

#end def Pedigree



#!/usr/bin/env python

#################################################################
#
#    validateVCF
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import json
import numpy
import copy
# matplotlib
from matplotlib import pyplot
# shared_functions as *
from granite.lib.shared_functions import *
# vcf_parser
from granite.lib import vcf_parser
# pedigree_parser
from granite.lib import pedigree_parser
# shared_vars
from granite.lib.shared_vars import real_NA_chroms


#################################################################
#
#    FUNCTIONS
#
#################################################################
#################################################################
#    Heterozygosity stats
#################################################################
def get_stats_het(vnt_obj, stat_dict, family, NA_chroms, var_types):
    ''' extract heterozygosity information from variant using family '''
    var_type = variant_type_ext(vnt_obj.REF, vnt_obj.ALT)
    if vnt_obj.CHROM.replace('chr', '') in NA_chroms:
        return # skip unbalanced chromosomes
    #end if
    is_ID = True if vnt_obj.ID != '.' else False
    if var_type in var_types:
        _error_het(vnt_obj, stat_dict, family, is_ID)
        _error_het_family(vnt_obj, stat_dict, family)
    #end if
#end def

def _error_het_family(vnt_obj, stat_dict, family):
    ''' calculate error model for heterozygous calls using family, MODEL '''
    # Counting children that are heterozygous
    cnt_children = 0
    for child in family['children']:
        try:
            GT_ = vnt_obj.get_genotype_value(child.sample, 'GT').replace('|', '/')
        except Exception: continue
        #end try
        if GT_ == './.' or GT_ == '1/1': return
        elif GT_ in ['0/1', '1/0']: cnt_children += 1
        #end if
    #end for
    # Parents samples
    sample_0 = family['parents'][0].sample
    sample_1 = family['parents'][1].sample
    # Update stat_dict for cnt_children
    stat_dict['error_het_family'].setdefault(cnt_children, {})
    for sample in [sample_0, sample_1]:
        stat_dict['error_het_family'][cnt_children].setdefault(sample, {
                                                            'ref': 0, # undercalled het
                                                            'hom': 0, # overcalled het
                                                            'het': 0, # properly called het
                                                            'missing': 0, # missing genotype
                                                            'total': 0
                                                        })
    #end for
    # Get trio genotypes for parents
    trio_0 = _GT_trio(vnt_obj, family['parents'][0])
    trio_1 = _GT_trio(vnt_obj, family['parents'][1])
    # Check trio genotypes combination
    if trio_0 and trio_1: # trio genotypes are complete
        if trio_0 == ['ref', {'ref'}] and trio_1[1] == {'ref', 'het'}:
            stat_dict['error_het_family'][cnt_children][sample_1][trio_1[0]] += 1
            stat_dict['error_het_family'][cnt_children][sample_1]['total'] += 1
        elif trio_1 == ['ref', {'ref'}] and trio_0[1] == {'ref', 'het'}:
            stat_dict['error_het_family'][cnt_children][sample_0][trio_0[0]] += 1
            stat_dict['error_het_family'][cnt_children][sample_0]['total'] += 1
        #end if
    #end if
#end def

def _error_het(vnt_obj, stat_dict, family, is_ID):
    ''' calculate error model for heterozygous calls using family, GENERAL '''
    # Counting children that are heterozygous or homozygous alternate
    cnt_children = 0
    for child in family['children']:
        try:
            GT_ = vnt_obj.get_genotype_value(child.sample, 'GT').replace('|', '/')
        except Exception: continue
        #end try
        if GT_ == './.': return
        # if GT_ == './.' or GT_ == '1/1': return
        elif GT_ in ['0/1', '1/0', '1/1']: cnt_children += 1
        # elif GT_ in ['0/1', '1/0']: cnt_children += 1
        #end if
    #end for
    # Parents samples
    sample_0 = family['parents'][0].sample
    sample_1 = family['parents'][1].sample
    # Update stat_dict for cnt_children
    stat_dict['error_het'].setdefault(cnt_children, {})
    for sample in [sample_0, sample_1]:
        stat_dict['error_het'][cnt_children].setdefault(sample, {
                                                    'no_spouse': {
                                                        'is_ID': {
                                                            'in_gparents': 0,
                                                            'no_gparents': 0,
                                                            'total': 0
                                                        },
                                                        'no_ID': {
                                                            'in_gparents': 0,
                                                            'no_gparents': 0,
                                                            'total': 0
                                                        }
                                                    },
                                                    'in_spouse': {
                                                        'is_ID': {
                                                            'in_gparents': 0,
                                                            'no_gparents': 0,
                                                            'total': 0
                                                        },
                                                        'no_ID': {
                                                            'in_gparents': 0,
                                                            'no_gparents': 0,
                                                            'total': 0
                                                        }
                                                    }
                                                })
    #end for
    # Get trio genotypes for parents
    trio_0 = _GT_trio(vnt_obj, family['parents'][0])
    trio_1 = _GT_trio(vnt_obj, family['parents'][1])
    key_ID = 'is_ID' if is_ID else 'no_ID'
    if trio_0 and trio_1:
        if trio_0[0] == 'het':
            key_sps = 'no_spouse' if trio_1 == ['ref', {'ref'}] else 'in_spouse'
            stat_dict['error_het'][cnt_children][sample_0][key_sps][key_ID]['total'] += 1
            if trio_0[1] != {'ref'}:
                stat_dict['error_het'][cnt_children][sample_0][key_sps][key_ID]['in_gparents'] += 1
            else:
                stat_dict['error_het'][cnt_children][sample_0][key_sps][key_ID]['no_gparents'] += 1
            #end if
        #end if
        if trio_1[0] == 'het':
            key_sps = 'no_spouse' if trio_0 == ['ref', {'ref'}] else 'in_spouse'
            stat_dict['error_het'][cnt_children][sample_1][key_sps][key_ID]['total'] += 1
            if trio_1[1] != {'ref'}:
                stat_dict['error_het'][cnt_children][sample_1][key_sps][key_ID]['in_gparents'] += 1
            else:
                stat_dict['error_het'][cnt_children][sample_1][key_sps][key_ID]['no_gparents'] += 1
            #end if
        #end if
    #end if
#end def

#################################################################
#    De novo stats
#################################################################
def get_stats_novo(vnt_obj, stat_dict, family, NA_chroms, var_types, sample_novo, novotag):
    ''' extract de novo information from variant using family '''
    var_type = variant_type_ext(vnt_obj.REF, vnt_obj.ALT)
    if vnt_obj.CHROM.replace('chr', '') in NA_chroms:
        return # skip unbalanced chromosomes
    #end if
    if var_type in var_types:
        try:
            novoPP = float(vnt_obj.get_tag_value(novotag))
        except Exception: return
        #end try
        if novoPP > 0:
            _error_novo_family(vnt_obj, stat_dict, family, sample_novo, novoPP)
        #end if
    #end if
#end def

def _error_novo_family(vnt_obj, stat_dict, family, sample_novo, novoPP):
    ''' calculate error model for de novo calls using family '''
    # Counting children that are heterozygous or homozygous alternate
    cnt_children, is_child = 0, False
    for child in family['children']:
        if child.sample != sample_novo:
            try:
                GT_ = vnt_obj.get_genotype_value(child.sample, 'GT').replace('|', '/')
            except Exception: continue
            #end try
            if GT_ == './.': return
            elif GT_ in ['0/1', '1/0', '1/1']: cnt_children += 1
            #end if
        else: is_child = True
        #end if
    #end for
    if novoPP >= 0.9: bin = 0.9
    elif novoPP >= 0.1: bin = 0.1
    else: bin = 0
    #end if
    # Update stat_dict for cnt_children
    stat_dict['error_novo_family'].setdefault(bin, {})
    stat_dict['error_novo_family'][bin].setdefault(cnt_children, {
                                                            'in_gparents': 0,
                                                            'no_gparents': 0,
                                                            'total': 0,
                                                            'no_gparents_vnt': [],
                                                            'total_vnt': [],
                                                            'no_gparents_AD-DP': []
                                                        })
    # Check if sample_novo is child or parent
    if is_child:
        _error_novo_child(vnt_obj, stat_dict, family, bin, cnt_children, sample_novo)
    else:
        _error_novo_parent(vnt_obj, stat_dict, family, bin, cnt_children, sample_novo)
    #end if
#end def

def _error_novo_child(vnt_obj, stat_dict, family, bin, cnt_children, sample_novo):
    ''' sample to calculate error for is child '''
    # Get trio genotypes for parents
    trio_0 = _GT_trio(vnt_obj, family['parents'][0])
    trio_1 = _GT_trio(vnt_obj, family['parents'][1])
    # Check trio genotypes combination
    if trio_0 and trio_1: # trio genotypes are complete
        stat_dict['error_novo_family'][bin][cnt_children]['total'] += 1
        stat_dict['error_novo_family'][bin][cnt_children]['total_vnt'].append(vnt_obj)
        if trio_0[1] != {'ref'} or trio_1[1] != {'ref'}:
            stat_dict['error_novo_family'][bin][cnt_children]['in_gparents'] += 1
        else:
            stat_dict['error_novo_family'][bin][cnt_children]['no_gparents'] += 1
            stat_dict['error_novo_family'][bin][cnt_children]['no_gparents_vnt'].append(vnt_obj)
            # Get AD-DP ratio
            AD = sum(map(int, vnt_obj.get_genotype_value(sample_novo, 'AD').split(',')[1:]))
            DP = int(vnt_obj.get_genotype_value(sample_novo, 'DP'))
            try:
                stat_dict['error_novo_family'][bin][cnt_children]['no_gparents_AD-DP'].append(AD / DP)
            except Exception:
                stat_dict['error_novo_family'][bin][cnt_children]['no_gparents_AD-DP'].append(0)
            #end try
        #end if
    #end if
#end def

def _error_novo_parent(vnt_obj, stat_dict, family, bin, cnt_children, sample_novo):
    ''' sample to calculate error for is parent '''
    # Get trio for parent only
    for parent in family['parents']:
        if parent.sample == sample_novo:
            trio = _GT_trio(vnt_obj, parent)
            break
        #end if
    #end for
    if trio: # ignoring spouse family, checking parent only
        stat_dict['error_novo_family'][bin][cnt_children]['total'] += 1
        stat_dict['error_novo_family'][bin][cnt_children]['total_vnt'].append(vnt_obj)
        if trio[1] != {'ref'}:
            stat_dict['error_novo_family'][bin][cnt_children]['in_gparents'] += 1
        else:
            stat_dict['error_novo_family'][bin][cnt_children]['no_gparents'] += 1
            stat_dict['error_novo_family'][bin][cnt_children]['no_gparents_vnt'].append(vnt_obj)
            # Get AD-DP ratio
            AD = sum(map(int, vnt_obj.get_genotype_value(sample_novo, 'AD').split(',')[1:]))
            DP = int(vnt_obj.get_genotype_value(sample_novo, 'DP'))
            stat_dict['error_novo_family'][bin][cnt_children]['no_gparents_AD-DP'].append(AD / DP)
        #end if
    #end if
#end def

#################################################################
#    Routines for stats
#################################################################
def _GT_trio(vnt_obj, member_obj):
    ''' return genotype information for member_obj and parents,
    [GT_member_obj, set([GT_parent_0, GT_parent_1])] '''
    encode = {'0/0': 'ref', '1/1': 'hom',
              '1/0': 'het', '0/1': 'het',
              './.': 'missing'}
    GT_trio, GT_parents = [], set()
    GT = vnt_obj.get_genotype_value(member_obj.sample, 'GT').replace('|', '/')
    for parent in member_obj.get_parents():
        GT_ = vnt_obj.get_genotype_value(parent.sample, 'GT').replace('|', '/')
        if GT_ == './.': return
        else: GT_parents.add(encode[GT_])
        #end if
    #end for
    GT_trio.append(encode[GT])
    GT_trio.append(GT_parents)
    return GT_trio
#end def

#################################################################
#
#################################################################
def to_json(stat_dict_list, sample_het_list, sample_novo):
    ''' '''
    stat_json = {}
    # Error heterozygous calls, model
    if sample_het_list:
        stat_json.setdefault('autosomal heterozygous calls error, MODEL', [])
        # Calculate error
        tmp_dict = {}
        for stat_dict in stat_dict_list:
            if stat_dict['error_het_family']:
                for cnt_children in sorted(stat_dict['error_het_family']):
                    tmp_dict.setdefault(cnt_children, {
                                                        'children': cnt_children
                                                        })
                    for sample in stat_dict['error_het_family'][cnt_children]:
                        if sample in sample_het_list:
                            tmp_dict[cnt_children].setdefault(sample, {})
                            _counts = {
                                'total': stat_dict['error_het_family'][cnt_children][sample]['total']
                            }
                            for k, v in stat_dict['error_het_family'][cnt_children][sample].items():
                                # Counts
                                if k == 'ref': _counts.setdefault('undercall', v)
                                elif k == 'hom': _counts.setdefault('overcall', v)
                                elif k == 'het': _counts.setdefault('correct', v)
                                elif k == 'missing': _counts.setdefault('missing', v)
                                #end if
                            #end for
                            # Ratios
                            if _counts['total']:
                                tmp_dict[cnt_children][sample].setdefault('undercall ratio', round(_counts['undercall'] / _counts['total'], 3))
                                tmp_dict[cnt_children][sample].setdefault('overcall ratio', round(_counts['overcall'] / _counts['total'], 3))
                                tmp_dict[cnt_children][sample].setdefault('correct ratio', round(_counts['correct'] / _counts['total'], 3))
                                tmp_dict[cnt_children][sample].setdefault('missing ratio', round(_counts['missing'] / _counts['total'], 3))
                            #end if
                            tmp_dict[cnt_children][sample].setdefault('counts', _counts)
                        #end if
                    #end for
                #end for
            #end if
        #end for
        for cnt_children in sorted(tmp_dict):
            stat_json['autosomal heterozygous calls error, MODEL'].append(tmp_dict[cnt_children])
        #end for
        _extend_json_het(stat_json)
    #end if
    # Error heterozygous calls, general
    if sample_het_list:
        stat_json.setdefault('autosomal heterozygous calls error, GENERAL', [])
        # Calculate error
        tmp_dict = {}
        for sample in sample_het_list:
            tmp_dict.setdefault(sample, {
                                'sample': sample,
                                'by_siblings': {}
                                })
        #end for
        for stat_dict in stat_dict_list:
            if stat_dict['error_het']:
                for cnt_children in sorted(stat_dict['error_het']):
                    for sample in stat_dict['error_het'][cnt_children]:
                        if sample in tmp_dict:
                            tmp_dict[sample]['by_siblings'].setdefault(cnt_children, {
                                                                            'in_gparents': 0,
                                                                            'no_gparents': 0,
                                                                            'total': 0,
                                                                            'is_ID': {
                                                                                'in_gparents': 0,
                                                                                'no_gparents': 0,
                                                                                'total': 0
                                                                            },
                                                                            'no_spouse': {
                                                                                'is_ID': {
                                                                                    'in_gparents': 0,
                                                                                    'no_gparents': 0,
                                                                                    'total': 0
                                                                                },
                                                                                'in_gparents': 0,
                                                                                'no_gparents': 0,
                                                                                'total': 0
                                                                            }
                                                                        })
                            # In spouse
                            for k, v in stat_dict['error_het'][cnt_children][sample]['in_spouse']['is_ID'].items():
                                tmp_dict[sample]['by_siblings'][cnt_children][k] += v
                                tmp_dict[sample]['by_siblings'][cnt_children]['is_ID'][k] += v
                            #end for
                            for k, v in stat_dict['error_het'][cnt_children][sample]['in_spouse']['no_ID'].items():
                                tmp_dict[sample]['by_siblings'][cnt_children][k] += v
                            #end for
                            # No spouse
                            for k, v in stat_dict['error_het'][cnt_children][sample]['no_spouse']['is_ID'].items():
                                tmp_dict[sample]['by_siblings'][cnt_children][k] += v
                                tmp_dict[sample]['by_siblings'][cnt_children]['is_ID'][k] += v
                                tmp_dict[sample]['by_siblings'][cnt_children]['no_spouse'][k] += v
                                tmp_dict[sample]['by_siblings'][cnt_children]['no_spouse']['is_ID'][k] += v
                            #end for
                            for k, v in stat_dict['error_het'][cnt_children][sample]['no_spouse']['no_ID'].items():
                                tmp_dict[sample]['by_siblings'][cnt_children][k] += v
                                tmp_dict[sample]['by_siblings'][cnt_children]['no_spouse'][k] += v
                            #end for
                        #end if
                    #end for
                #end for
            #end if
        #end for
        for sample in sorted(tmp_dict):
            stat_json['autosomal heterozygous calls error, GENERAL'].append(tmp_dict[sample])
        #end for
    #end if
    # Error de novo calls
    if sample_novo:
        stat_json.setdefault('de novo calls error', [])
        for stat_dict in stat_dict_list: # only one stat_dict at most
            if stat_dict['error_novo_family']:
                tmp_dict = {
                        'novoPP': '',
                        'total variants': 0,
                        'not in grandparents': 0,
                        'not in grandparents, in siblings': {}
                        }
                for bin in sorted(stat_dict['error_novo_family'], reverse=True):
                    if bin != 0: tmp_dict['novoPP'] = '>= ' + str(bin)
                    else: tmp_dict['novoPP'] = '> ' + str(bin)
                    #end if
                    for cnt_children in sorted(stat_dict['error_novo_family'][bin]):
                        if stat_dict['error_novo_family'][bin][cnt_children]['no_gparents']:
                            tmp_dict['not in grandparents, in siblings'].setdefault(cnt_children, 0)
                        #end if
                        for k, v in stat_dict['error_novo_family'][bin][cnt_children].items():
                            if k == 'total':
                                tmp_dict['total variants'] += v
                            elif k == 'no_gparents' and v:
                                tmp_dict['not in grandparents'] += v
                                tmp_dict['not in grandparents, in siblings'][cnt_children] += v
                            #end if
                        #end for
                    #end for
                    stat_json['de novo calls error'].append(copy.deepcopy(tmp_dict))
                #end for
            #end if
        #end for
    #end if
    return stat_json
#end def

def _extend_json_het(stat_json):
    ''' '''
    # Error heterozygous calls in cnt_children range
    cnt_children = stat_json['autosomal heterozygous calls error, MODEL'][-1]['children']
    median = cnt_children // 2 + 1
    range = cnt_children * 25 // 100
    min = median - range
    max = median + range
    dict_range = {
        'children': '[' + str(min) + ',' + str(max) + ']',
        'counts': {
          "undercall": 0, "overcall": 0,
          "correct": 0, "missing": 0, "total": 0
        }
    }
    dict_median = {
        'children': '[' + str(median) + ',' + str(cnt_children) + ']',
        'counts': {
          "undercall": 0, "overcall": 0,
          "correct": 0, "missing": 0, "total": 0
        }
    }
    for tmp_dict in stat_json['autosomal heterozygous calls error, MODEL']:
        cnt_children_ = tmp_dict['children']
        if cnt_children_ >= min:
            for k, v in tmp_dict.items():
                if k != 'children':
                    for k_v, v_v in v['counts'].items():
                        if cnt_children_ <= max:
                            dict_range['counts'][k_v] += v_v
                        #end if
                        if cnt_children_ >= median:
                            dict_median['counts'][k_v] += v_v
                        #end if
                    #end for
                #end if
            #end for
        #end if
    #end for
    for dict_ in [dict_range, dict_median]:
        if dict_['counts']['total']:
            dict_.setdefault('undercall ratio', round(dict_['counts']['undercall'] / dict_['counts']['total'], 3))
            dict_.setdefault('overcall ratio', round(dict_['counts']['overcall'] / dict_['counts']['total'], 3))
            dict_.setdefault('correct ratio', round(dict_['counts']['correct'] / dict_['counts']['total'], 3))
            dict_.setdefault('missing ratio', round(dict_['counts']['missing'] / dict_['counts']['total'], 3))
        #end if
        stat_json['autosomal heterozygous calls error, MODEL'].append(dict_)
    #end for
#end def

def novo_variants(stat_dict_list, filename):
    ''' '''
    for stat_dict in stat_dict_list: # only one stat_dict at most
        if stat_dict['error_novo_family']:
            fv = open(filename.split('.')[0] + '_total_variants.txt', 'w')
            for bin in sorted(stat_dict['error_novo_family'], reverse=True):
                for cnt_children in sorted(stat_dict['error_novo_family'][bin]):
                    if stat_dict['error_novo_family'][bin][cnt_children]['total_vnt']:
                        with open(filename.split('.')[0] + '_total_' + str(bin) + '-' + str(cnt_children) + '.txt', 'w') as fo:
                            for vnt_obj in stat_dict['error_novo_family'][bin][cnt_children]['total_vnt']:
                                fo.write('{0}\t{1}\t{2}\t{3}\n'.format(vnt_obj.CHROM, vnt_obj.POS, vnt_obj.REF, vnt_obj.ALT))
                                fv.write(vnt_obj.to_string())
                            #end for
                        #end with
                    #end if
                    if stat_dict['error_novo_family'][bin][cnt_children]['no_gparents_vnt']:
                        with open(filename.split('.')[0] + '_no-gprnts_' + str(bin) + '-' + str(cnt_children) + '.txt', 'w') as fo:
                            for vnt_obj in stat_dict['error_novo_family'][bin][cnt_children]['no_gparents_vnt']:
                                fo.write('{0}\t{1}\t{2}\t{3}\n'.format(vnt_obj.CHROM, vnt_obj.POS, vnt_obj.REF, vnt_obj.ALT))
                            #end for
                        #end with
                    #end if
                #end for
            #end for
            fv.close()
        #end if
    #end for
#end def

#################################################################
#    Plots
#################################################################
def plot_error_het_family(stat_dict_list, sample_het_list):
    ''' plot accuracy for autosomal heterozygous variants in parents,
    family '''
    # Labels
    sample_0, sample_1 = sorted(sample_het_list)
    # Data
    data = {
        'bins': [],
        sample_0: [],
        sample_1: []
    }
    # Get data points
    for stat_dict in stat_dict_list:
        if stat_dict['error_het_family']:
            for cnt_children in sorted(stat_dict['error_het_family']):
                if cnt_children not in data['bins']:
                    data['bins'].append(cnt_children)
                #end if
                for sample in stat_dict['error_het_family'][cnt_children]:
                    if sample in sample_het_list:
                        try:
                            correct = stat_dict['error_het_family'][cnt_children][sample]['het'] / \
                                      stat_dict['error_het_family'][cnt_children][sample]['total']
                            data[sample].append(correct * 100)
                        except Exception: data[sample].append(0)
                        #end try
                    #end if
                #end for
            #end for
        #end if
    #end for
    filename = 'autosomal_heterozygous_accuracy_family_{0}-{1}.png'.format(sample_0, sample_1)
    xlabel = 'Number of heterozygous children'
    ylabel = '% of correct calls in parent'
    title = 'Accuracy of autosomal heterozygous variant calls (family)'
    _plot_hist_2(data['bins'], data[sample_0], data[sample_1],
                        sample_0, sample_1, filename, xlabel, ylabel, title, percent=True)
#end def

def plot_distr_het_family(stat_dict_list, sample_het_list):
    ''' plot distribution for autosomal heterozygous variants in parents,
    family '''
    # Labels
    sample_0, sample_1 = sorted(sample_het_list)
    # Data
    data = {
        'bins': [],
        sample_0: [],
        sample_1: []
    }
    # Get data points
    for stat_dict in stat_dict_list:
        if stat_dict['error_het_family']:
            for cnt_children in sorted(stat_dict['error_het_family']):
                if cnt_children not in data['bins']:
                    data['bins'].append(cnt_children)
                #end if
                for sample in stat_dict['error_het_family'][cnt_children]:
                    if sample in sample_het_list:
                        try:
                            correct = stat_dict['error_het_family'][cnt_children][sample]['het']
                            data[sample].append(correct)
                        except Exception: data[sample].append(0)
                        #end try
                    #end if
                #end for
            #end for
        #end if
    #end for
    filename = 'autosomal_heterozygous_distribution_family_{0}-{1}.png'.format(sample_0, sample_1)
    xlabel = 'Number of heterozygous children'
    ylabel = 'Number of variants heterozygous in parent'
    title = 'Distribution of autosomal heterozygous variant calls (family)'
    _plot_hist_2(data['bins'], data[sample_0], data[sample_1],
                        sample_0, sample_1, filename, xlabel, ylabel, title)
#end def

def plot_distr_het(stat_dict_list, sample_het_list):
    ''' plot distribution for autosomal heterozygous variants in parents,
    trio '''
    # Labels
    sample_0, sample_1 = sorted(sample_het_list)
    # Data
    data = {
        'bins': [],
        sample_0: [],
        sample_1: []
    }
    # Get data points
    for stat_dict in stat_dict_list:
        if stat_dict['error_het']:
            for cnt_children in sorted(stat_dict['error_het']):
                if cnt_children not in data['bins']:
                    data['bins'].append(cnt_children)
                #end if
                for sample in stat_dict['error_het'][cnt_children]:
                    if sample in sample_het_list:
                        try:
                            data[sample].append(stat_dict['error_het'][cnt_children][sample]['no_spouse']['is_ID']['in_gparents'])
                        except Exception: data[sample].append(0)
                        #end try
                    #end if
                #end for
            #end for
        #end if
    #end for
    filename = 'autosomal_heterozygous_distribution_trio_{0}-{1}.png'.format(sample_0, sample_1)
    xlabel = 'Number of heterozygous children'
    ylabel = 'Number of variants heterozygous in parent'
    title = 'Distribution of autosomal heterozygous variant calls (trio)'
    _plot_hist_2(data['bins'], data[sample_0], data[sample_1],
                        sample_0, sample_1, filename, xlabel, ylabel, title)
#end def

def plot_AD_DP_ratio(stat_dict, sample_novo):
    ''' '''
    data_0, data_no_0 = [], []
    for cnt_children in stat_dict['error_novo_family'][0.9]:
        if cnt_children == 0:
            data_0 += stat_dict['error_novo_family'][0.9][cnt_children]['no_gparents_AD-DP']
        else:
            data_no_0 += stat_dict['error_novo_family'][0.9][cnt_children]['no_gparents_AD-DP']
        #end if
    #end for
    # Plotting
    pyplot.hist(numpy.array(data_0), alpha=0.5, color='#67a9cf', label='in 0 children')
    pyplot.hist(numpy.array(data_no_0), alpha=0.5, color='#ef8a62', label='in 1+ children')
    # Format and save
    pyplot.gca().set(title='AD-DP ratio distribution (novoPP >= 0.9)', ylabel='Frequency')
    pyplot.legend()
    filename = 'AD-DP_distribution_denovo_{0}.png'.format(sample_novo)
    pyplot.savefig(filename)
#end def

def _plot_hist_2(bins, data_0, data_1, label_0, label_1, filename, xlabel, ylabel, title, percent=False):
    ''' '''
    # Create plot
    fig, ax = pyplot.subplots()
    index = numpy.arange(len(bins))
    bar_width = 0.35
    # Plot data
    rects1 = pyplot.bar(index, data_0, bar_width, color='#ef8a62', label=label_0, zorder=3)
    rects2 = pyplot.bar(index + bar_width + 0.02, data_1, bar_width, color='#67a9cf', label=label_1, zorder=3)
    # Grid and axis
    pyplot.grid(axis='y', alpha=0.8, zorder=0, linestyle='--')
    pyplot.xticks(index + bar_width / 2 + 0.01, bins)
    if percent: pyplot.ylim(0, 120)
    #end if
    yticks = ax.yaxis.get_major_ticks()
    yticks[-1].set_visible(False)
    # Labels and layout
    pyplot.xlabel(xlabel)
    pyplot.ylabel(ylabel)
    pyplot.title(title)
    pyplot.legend()
    pyplot.tight_layout()
    # Save
    pyplot.savefig(filename)
#end def

#################################################################
#    runner
#################################################################
def main(args):
    ''' '''
    # Variables
    NA_chroms = real_NA_chroms
    is_verbose = True if args['verbose'] else False
    sample_novo = args['novo'] if args['novo'] else ''
    sample_het_list = args['het'] if args['het'] else []
    anchor_list = args['anchor']
    var_types = [s.lower() for s in args['type']] if args['type'] else ['snv']

    # Check pedigree and anchor args
    if len(anchor_list) != len(args['pedigree']) and \
       len(args['pedigree']) != 1:
        sys.exit('\nERROR in parsing arguments: number of pedigrees and anchors is different\n')
    #end if

    # Buffers
    fo = open(args['outputfile'], 'w')

    # Creating Vcf object
    vcf_obj = vcf_parser.Vcf(args['inputfile'])

    # Get pedigree / pedigrees information
    pedigree_list = []
    for pedigree in args['pedigree']:
        # Loading pedigree
        if os.path.isfile(pedigree):
            with open(pedigree) as fi:
                pedigree_list.append(json.load(fi))
            #end with
        else:
            try: pedigree_list.append(json.loads(pedigree))
            except Exception:
                sys.exit('\nERROR in parsing arguments: {0} must be either a json file or a string representing a json\n'
                        .format(pedigree))
            #end try
        #end if
    #end for

    # Check novoPP
    if sample_novo:
        try: novotag, _ = vcf_obj.header.check_tag_definition('novoPP')
        except Exception:
                sys.exit('\nERROR in parsing arguments: novoCaller information missing in VCF file\n')
        #end try
    #end if

    # Creating Pedigree object / objects
    pedigree_obj_list = []
    for pedigree in pedigree_list:
        pedigree_obj_list.append(pedigree_parser.Pedigree(pedigree))
    #end for

    # Initializing stat_dict for each anchor
    stat_dict_list = []
    for anchor in anchor_list:
        stat_dict_list.append({
                            'error_het_family': {},
                            'error_het': {},
                            'error_novo_family': {}
                             })
    #end for

    # Building family / families
    family_list = []
    if len(pedigree_obj_list) == 1:
        for anchor in anchor_list:
            family_list.append(pedigree_obj_list[0].get_family(anchor))
        #end for
    else:
        for i, pedigree_obj in enumerate(pedigree_obj_list):
            family_list.append(pedigree_obj.get_family(anchor_list[i]))
        #end for
    #end if

    # Reading variants
    analyzed = 0
    vnt_obj_ = None
    for i, vnt_obj in enumerate(vcf_obj.parse_variants()):
        if is_verbose:
            sys.stderr.write('\rAnalyzing variant... ' + str(i + 1))
            sys.stderr.flush()
        #end if

        # # Check if chromosome is canonical and in valid format
        # if not check_chrom(vnt_obj.CHROM):
        #     continue
        # #end if
        analyzed += 1

        # Skip MAV that are redundant
        if vnt_obj_:
            if vnt_obj_.CHROM == vnt_obj.CHROM and vnt_obj_.POS == vnt_obj.POS:
                continue
            #end if
        #end if
        vnt_obj_ = vnt_obj

        # Getting and updating stats for each stat_dict / family
        for l, family in enumerate(family_list):
            if sample_novo:
                if anchor_list[l] == sample_novo:
                    get_stats_novo(vnt_obj, stat_dict_list[l], family, NA_chroms, var_types, sample_novo, novotag)
                #end if
            #end if
            if sample_het_list:
                if anchor_list[l] in sample_het_list:
                    get_stats_het(vnt_obj, stat_dict_list[l], family, NA_chroms, var_types)
                #end if
            #end if
        #end for
    #end for

    # Writing output
    sys.stderr.write('\n\n...Writing results for ' + str(analyzed) + ' analyzed variants out of ' + str(i + 1) + ' total variants\n')
    sys.stderr.flush()

    # Create json
    stat_json = to_json(stat_dict_list, sample_het_list, sample_novo)

    # Write variants
    if sample_novo:
        novo_variants(stat_dict_list, args['outputfile'])
        for l, family in enumerate(family_list):
            if anchor_list[l] == sample_novo:
                try: plot_AD_DP_ratio(stat_dict_list[l], sample_novo)
                except Exception: # no variants with novoPP >= 0.9, skip
                    pass
                #end try
            #end if
        #end for
    #end if

    # Plots
    if len(sample_het_list) == 2:
        plot_error_het_family(stat_dict_list, sample_het_list)
        plot_distr_het_family(stat_dict_list, sample_het_list)
        plot_distr_het(stat_dict_list, sample_het_list)
    #end if

    # Write json to file
    json.dump(stat_json, fo, indent=2, sort_keys=False)

    # Closing buffers
    fo.close()
#end def


#################################################################
#
#    MAIN
#
#################################################################
if __name__ == "__main__":

    main()

#end if



#!/usr/bin/env python

# Libraries
from setuptools import setup, find_packages

# Description
with open('README.md') as fd:
    long_description = fd.read()
#end with

# Requirements
with open('requirements.txt') as fr:
    set_parsed = fr.read().splitlines()
#end with

# Set requires
install_requires = [req.strip() for req in set_parsed]
tests_requires = [
    'pytest'
]

# Setup
setup(
    name = 'granite-suite',
    version = open('granite/_version.py').readlines()[-1].split()[-1].strip("\"'"),
    description = 'granite is a collection of software to call, filter and work with genomic variants',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/dbmi-bgm/granite',
    classifiers=[
            'License :: OSI Approved :: MIT License',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3'
            ],
    install_requires=install_requires,
    setup_requires=install_requires,
    tests_require=tests_requires,
    include_package_data=True,
    packages = find_packages(),
    author = 'Michele Berselli',
    author_email = 'berselli.michele@gmail.com',
    license = 'MIT',
    python_requires = '>=3.6',
    entry_points = {
        'console_scripts': [
            'granite = granite.granite:main',
        ]
    }
    )



#!/usr/bin/env python

#################################################################
#
#    mpileup_parser
#        Michele Berselli
#        Harvard Medical School
#        berselli.michele@gmail.com
#
#################################################################


#################################################################
#
#    LIBRARIES
#
#################################################################
import sys, os
import re


#################################################################
#
#    mpileupParser
#      -> mpileupColumn
#
#################################################################
class mpileupParser(object):
    ''' class to manipulate mpileup format from samtools '''

    def __init__(self):
        self.mpileupColumns = []
    #end def

    class mpileupColumn(object):
        ''' class to manipulate mpileup column at position '''

        def __init__(self, chr, pos, ref, cov, reads, BQs):
            ''' '''
            self.chr = chr
            self.pos = int(pos)
            self.ref = ref
            self.cov = cov
            self.reads = reads
            self.BQs = BQs
            self.counts = {
                    'ref_fw': 0, 'alt_fw': 0, 'ref_rv': 0, 'alt_rv': 0,
                    'ins_fw': 0, 'ins_rv': 0, 'del_fw': 0, 'del_rv': 0
                    }
        #end def

        def __parser_reads(self, basic=True):
            ''' parse the reads field and return a list of the reads.
            can provide only basic info (e.g. match/mismatch/indel),
            or be more comprehensive and return all info for each read '''
            reads_list, reads = [], self.reads
            i, l_i = 0, 0 # l_i is current index in reads_list
            while i < len(reads):
                if reads[i] in "ACGTNacgtn.,*":
                    reads_list.append(reads[i])
                    i += 1; l_i += 1
                elif reads[i] == '$':
                    if not basic: reads_list[l_i-1] += '$'
                    #end if
                    i += 1
                elif reads[i] == '^':
                    if not basic: reads_list.append(reads[i] + reads[i+2])
                    else: reads_list.append(reads[i+2])
                    #end if
                    i += 3; l_i += 1
                elif reads[i] in '+-':
                    re_match = re.search('[\+-](\d+)([ACGTNacgtn*]+)', reads[i:])
                    indl_len = int(re_match.group(1))
                    if not basic: reads_list[l_i-1] += reads[i] + re_match.group(2)[:indl_len]
                    else : reads_list[l_i-1] += reads[i]
                    #end if
                    i += 1 + len(re_match.group(1)) + indl_len
                else:
                    raise ValueError('\nERROR in mpileup parser, unknown char {0} in {1}\n'
                              .format(reads[i], reads))
                #end if
            #end while
            return reads_list
        #end def

        def get_AD_noreference(self, ref):
            ''' return counts from reads '''
            encode = {
                    ref.upper(): 'ref_fw', ref.lower(): 'ref_rv',
                    'A+': 'ins_fw', 'A-': 'del_fw', 'a+': 'ins_rv', 'a-': 'del_rv',
                    'C+': 'ins_fw', 'C-': 'del_fw', 'c+': 'ins_rv', 'c-': 'del_rv',
                    'T+': 'ins_fw', 'T-': 'del_fw', 't+': 'ins_rv', 't-': 'del_rv',
                    'G+': 'ins_fw', 'G-': 'del_fw', 'g+': 'ins_rv', 'g-': 'del_rv',
                    'N+': 'ins_fw', 'N-': 'del_fw', 'n+': 'ins_rv', 'n-': 'del_rv'
                    }
            for read in self.__parser_reads():
                if '*' not in read:
                    try:
                        if read in encode:
                            self.counts[encode[read]] += 1
                        else:
                            if read.isupper(): self.counts['alt_fw'] += 1
                            else: self.counts['alt_rv'] += 1
                            #end if
                        #end if
                    except Exception:
                        raise ValueError('\nERROR in mpileup parser, unknown char {0} in reads\n'
                                  .format(read))
                    #end try
                #end if
            #end for
        #end def

        def __write_AD_header(self, fo):
            ''' '''
            fo.write('#chr\tpos\tcov\t')
            fo.write('ref_fw\tref_rv\talt_fw\talt_rv\t')
            fo.write('ins_fw\tins_rv\tdel_fw\tdel_rv\n')
        #end def

        def write_AD(self, fo, header=False):
            ''' '''
            if header:
                self.__write_AD_header(fo)
            #end if
            fo.write('{0}\t{1}\t{2}\t'.format(self.chr, self.pos, self.cov))
            fo.write('{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\n'
            .format(
              self.counts['ref_fw'], self.counts['ref_rv'],
              self.counts['alt_fw'], self.counts['alt_rv'],
              self.counts['ins_fw'], self.counts['ins_rv'],
              self.counts['del_fw'], self.counts['del_rv']
              ))
        #end def

    #end class mpileupColumn

    def generator(self, fi):
        ''' parse mpileup format and return mpileupColumn objects as generator '''
        for line in fi:
            try: line = str(line, 'utf-8').rstrip()
            except Exception: line = line.rstrip()
            #end try
            chr, pos, ref, cov, reads, BQs = line.split()
            yield self.mpileupColumn(chr, pos, ref, cov, reads, BQs)
        #end for
    #end def

#end class mpileupParser
